import json


class DockerCliClient(object):

    def __init__(self, cluster):

        control_nodes = cluster.get_machines(machine_type='control',
                                             machine_status='Ready')
        if not control_nodes:
            raise Exception("No control nodes in 'Ready' status found "
                            "in the cluster {0}".format(cluster.name))
        self.ctl = control_nodes[0]

        self.machines = cluster.get_machines()

    @staticmethod
    def _decode(res):
        """Decode docker output in json format

        Docker encodes into json separated lines which are
        many separated json objects instead of a single list.
        """

        data = [
            json.loads(line)
            for line in res.stdout_str.splitlines()
        ]
        return data

    def _get_machine(self, machine_name):
        for machine in self.machines:
            if machine.name == machine_name:
                return machine
        raise Exception("Docker node {0} not found in nodes {1}"
                        .format(machine_name, [m.name for m in self.machines]))

    def ps(self, machine_name):
        machine = self._get_machine(machine_name)
        res = machine.run_cmd("docker ps --format '{{json .}}'",
                              verbose=False, check_exit_code=True)
        return self._decode(res)

    def node_ls(self):
        res = self.ctl.run_cmd("docker node ls --format '{{json .}}'",
                               verbose=False, check_exit_code=True)
        return self._decode(res)

    def node_inspect(self, machine_name):
        res = self.ctl.run_cmd(f"docker node inspect {machine_name}"
                               " --format '{{json .}}'",
                               verbose=False, check_exit_code=True)
        return self._decode(res)

    def service_ls(self):
        res = self.ctl.run_cmd("docker service ls --format '{{json .}}'",
                               verbose=False, check_exit_code=True)
        return self._decode(res)

    def service_ps(self, service_name):
        res = self.ctl.run_cmd(f"docker service ps {service_name}"
                               " --format '{{json .}}'",
                               verbose=False, check_exit_code=True)
        return self._decode(res)

    def service_inspect(self, service_name):
        res = self.ctl.run_cmd(f"docker service inspect {service_name}"
                               " --format '{{json .}}'",
                               verbose=False, check_exit_code=True)
        return self._decode(res)

    def service_logs(self, service_name):
        res = self.ctl.run_cmd(f"docker service logs {service_name}"
                               " --no-trunc",
                               verbose=False, check_exit_code=True)
        return self._decode(res)

    def container_ls(self, machine_name):
        machine = self._get_machine(machine_name)
        res = machine.run_cmd("docker container ls --format '{{json .}}'",
                              verbose=False, check_exit_code=True)
        return self._decode(res)

    def container_inspect(self, machine_name, container_name):
        machine = self._get_machine(machine_name)
        res = machine.run_cmd(f"docker container inspect {container_name}"
                              " --format '{{json .}}'",
                              verbose=False, check_exit_code=True)
        return self._decode(res)

    def container_logs(self, machine_name, container_name):
        machine = self._get_machine(machine_name)
        res = machine.run_cmd(f"docker container logs {container_name}",
                              verbose=False, check_exit_code=True)
        return self._decode(res)

    def stats(self):
        res = self.ctl.run_cmd("docker stats --no-stream --no-trunc"
                               " --format '{{json .}}'",
                               verbose=False, check_exit_code=True)
        return self._decode(res)
