#    Copyright 2017 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations


import yaml
import requests
import os
from retry import retry

from si_tests import logger
from kubernetes.client.rest import ApiException
from urllib3.exceptions import MaxRetryError

LOG = logger.logger


class K8sBaseResource(object):
    resource_type = None

    def __init__(self, manager, name=None, namespace=None, data=None):
        self._manager = manager
        self._name = name
        self._namespace = namespace
        self._read_cache = None
        if data is not None:
            self._update_cache(data)

    def __repr__(self):
        uid = 'unknown-uid'
        if self._read_cache is not None:
            uid = self.uid
        return "<{0} name='{1}' namespace='{2}' uuid='{3}'>".format(
            self.__class__.__name__, self.name, self.namespace, uid)

    @property
    def name(self):
        return self._name

    @property
    def namespace(self):
        return self._namespace

    @property
    def uid(self):
        return self.read(cached=True).metadata.uid

    def _update_cache(self, data):
        self._read_cache = data
        LOG.debug("Read: {}".format(yaml.dump(data)))
        self._namespace = data.metadata.namespace
        self._name = data.metadata.name

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def read(self, cached=False, **kwargs):
        if not cached or self._read_cache is None:
            self._update_cache(self._read(**kwargs))
        return self._read_cache

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def create(self, body, **kwargs):
        if isinstance(body, dict):
            body = self._manager._prepare_dict(body)
        LOG.info("K8S API Creating {0} with body:\n{1}".format(
                 self.resource_type, yaml.dump(body)))
        self._update_cache(self._create(body, **kwargs))
        return self

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def patch(self, body, **kwargs):
        if isinstance(body, dict):
            body = self._manager._prepare_dict(body)
        LOG.info("K8S API Patching {0} name={1} ns={2} with "
                 "body:\n{3}".format(self.resource_type, self.name,
                                     self.namespace, yaml.dump(body)))
        self._update_cache(self._patch(body, **kwargs))
        return self

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def replace(self, body, **kwargs):
        if isinstance(body, dict):
            body = self._manager._prepare_dict(body)
        LOG.debug("K8S API Replacing {0} name={1} ns={2} with "
                  "body:\n{3}".format(self.resource_type, self.name,
                                      self.namespace, yaml.dump(body)))
        self._update_cache(self._replace(body, **kwargs))
        return self

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def delete(self, **kwargs):
        self._delete(**kwargs)
        return self

    def __eq__(self, other):
        if not isinstance(other, K8sBaseResource):
            return NotImplemented

        if not isinstance(other, self.__class__):
            return False

        return self.uid == other.uid


class K8sBaseManager(object):
    resource_class = None

    def __init__(self, cluster):
        self._cluster = cluster

    @property
    def resource_type(self):
        return self.resource_class.resource_type

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def get(self, name=None, namespace=None, data=None):
        namespace = namespace or self._cluster.default_namespace
        return self.resource_class(self, name, namespace, data)

    def create(self, name=None, namespace=None, body=None, **kwargs):
        return self.get(name=name, namespace=namespace).create(body, **kwargs)

    def update(self, name=None, namespace=None, body=None, **kwargs):
        return self.get(name=name, namespace=namespace).patch(body, **kwargs)

    def __resource_from_data(self, data):
        return self.resource_class(self, data=data)

    def __list_filter(self, items, name_prefix=None):
        if name_prefix is not None:
            items = [item for item in items if
                     item.metadata.name.startswith(name_prefix)]
        return items

    def __list_to_resource(self, items):
        return [self.__resource_from_data(item) for item in items]

    def _deserialize_custom_object(self, data, object_klass):
        return self.api.api_client._ApiClient__deserialize(
            data, object_klass)

    def _deserialize_custom_object_list(self, data, list_klass, object_klass):
        result_list = self.api.api_client._ApiClient__deserialize(
            data, list_klass)
        result_items = [
            self._deserialize_custom_object(item, object_klass)
            for item in result_list.items]
        result_list.items = result_items
        return result_list

    def _prepare_dict(self, data):
        # After getting k8s objects using .to_dict(), the resulting dict
        # cannot be used in create or patch, because of difference
        # in the attributes from `attribute_map`.
        # Convert attribute name to json key in model definition for request.
        # Keys that are missing in `attribute_map` will be kept as is.
        attribute_map = {
            'api_version': 'apiVersion',
        }
        return {attribute_map.get(attr, attr): value
                for attr, value in data.items()}

    def list(self, namespace=None, name_prefix=None, **kwargs):
        namespace = namespace or self._cluster.default_namespace
        items = self.list_raw(namespace=namespace, **kwargs).items
        items = self.__list_filter(items, name_prefix=name_prefix)
        return self.__list_to_resource(items)

    def list_all(self, name_prefix=None, **kwargs):
        items = self.list_raw(**kwargs).items
        items = self.__list_filter(items, name_prefix=name_prefix)
        return self.__list_to_resource(items)

    @retry((ApiException, MaxRetryError), delay=10, tries=3, logger=LOG)
    def list_raw(self, namespace=None, **kwargs):
        """A special method to return raw data -
           full info about listed object
           Returns object like V1PodList (not K8sPod list)
        """
        if namespace:
            return self._list(namespace=namespace, **kwargs)
        else:
            return self._list_all(**kwargs)


def read_yaml_str(yaml_str):
    """ load yaml from string helper """
    return yaml.safe_load(yaml_str)


def read_yaml_file(file_path, *args):
    """ load yaml from joined file_path and *args helper """
    with open(os.path.join(file_path, *args)) as f:
        return yaml.safe_load(f)


def read_yaml_url(yaml_file_url):
    """ load yaml from url helper """
    return yaml.safe_load(requests.get(yaml_file_url).text)
