#    Copyright 2017 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations

import datetime
import exec_helpers
from kubernetes.client.rest import ApiException
from websocket._exceptions import WebSocketBadStatusException
from urllib3.exceptions import MaxRetryError
from kubernetes.stream import stream
from retry import retry
import os

from si_tests import settings
import si_tests.utils.waiters as helpers
from si_tests.clients.k8s.base import K8sBaseResource
from si_tests.clients.k8s.base import K8sBaseManager
from si_tests import logger
import yaml

LOG = logger.logger


class K8sPod(K8sBaseResource):
    resource_type = 'pod'

    def _read(self, **kwargs):
        return self._manager.api.read_namespaced_pod(
            self.name, self.namespace, **kwargs)

    def _create(self, body, **kwargs):
        return self._manager.api.create_namespaced_pod(
            self.namespace, body, **kwargs)

    def _patch(self, body, **kwargs):
        return self._manager.api.patch_namespaced_pod(
            self.name, self.namespace, body, **kwargs)

    def _replace(self, body, **kwargs):
        return self._manager.api.replace_namespaced_pod(
            self.name, self.namespace, body, **kwargs)

    def _delete(self, async_del=False, timeout=120, **kwargs):
        uid = self.uid
        self._manager.api.delete_namespaced_pod(
            self.name, self.namespace, **kwargs)
        if not async_del:
            helpers.wait(lambda: not self.exists(uid),
                         timeout=timeout,
                         timeout_msg='Timeout deleting pod {0}/{1}'
                         .format(self.namespace, self.name))

    def exists(self, uid=None):
        if not uid:
            uid = self.uid
        try:
            if self._read().to_dict()['metadata']['uid'] == uid:
                return True
            else:
                return False
        except ApiException as e:
            if e.status == 404:
                return False
            raise e

    def cp_from_pod(self, source_dir, source_file='.',
                    destination=settings.ARTIFACTS_DIR):
        """Method that provides kubectl cp functionality from pod to localhost.

        :param source_dir: directory to use as a root of the archive
        :param source_file: file/directory name or mask to archive and copy
        :param destination: local directory to store files and directories

        Basically it archives source using kubectl exec, write binary
        stdout to local tar file using base64 and unpack it to destination dir.
        """
        pod_name = self.read().metadata.name

        timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
        tmp_tar_base64 = (
            '{path}test_{pod_name}_{timestamp}.tar.base64'
            .format(path=destination,
                    pod_name=pod_name,
                    timestamp=timestamp))

        LOG.info("Copy files from the pod {0}/{1}:{2}/{3} "
                 "to the local directory {4}"
                 .format(self.name, self.namespace, source_dir,
                         source_file, destination))

        if source_file == '.':
            exists = self.check_non_empty_folder(source_dir)
        else:
            exists = self.check_file_exists(os.path.join(source_dir,
                                                         source_file))

        if not exists:
            LOG.error("There is nothing to copy")
            return None

        # Command for pack and code files to tar.base64
        exec_command = ['/bin/sh', '-c',
                        'cd {0}; tar -C {0} -c {1} | base64'
                        .format(source_dir, source_file)]

        with open(tmp_tar_base64, "w") as tar_buffer:

            resp = stream(self._manager.api.connect_get_namespaced_pod_exec,
                          self.name, self.namespace,
                          command=exec_command,
                          stderr=True, stdin=True,
                          stdout=True, tty=False,
                          _preload_content=False,
                          _request_timeout=1800)

            while resp.is_open():
                resp.update(timeout=1)
                if resp.peek_stdout():
                    out = resp.read_stdout()
                    tar_buffer.write(out)
                if resp.peek_stderr():
                    LOG.error("STDERR: %s" % resp.read_stderr())
            resp.close()

            tar_buffer.flush()
            tar_buffer.seek(0)

        # Decode and unpack files from tar.base64 to 'destination'
        exec_helpers.Subprocess().check_call(
            "base64 -d {0} | tar -C {1} -x"
            .format(tmp_tar_base64, destination))
        # Remove tar.base64 file
        exec_helpers.Subprocess().check_call(
            "rm {0}".format(tmp_tar_base64))

    def check_file_exists(self, filepath=''):
        res = self.exec(
            ['/bin/sh', '-c',
             'test -f {0} && echo "1" || echo "0"'.format(
                 filepath)]).strip()
        return True if res != "0" else False

    def check_folder_exists(self, dir_path=''):
        """Check if directory in pod exists
        """
        res = self.exec(
            ['/bin/sh', '-c',
             'test -d {0}/ && echo "1" || echo "0"'.format(
                 dir_path)]).strip()
        return True if res != "0" else False

    def check_non_empty_folder(self, path=''):
        """Check if folder in pod is non-empty
        """
        res = self.exec(
            ['/bin/sh', '-c',
             'test -d {0}/ && ls -A {0} | wc -l || echo "0"'.format(
                 path)]).strip()
        return True if res != "0" else False

    @retry((ApiException, MaxRetryError, WebSocketBadStatusException),
           delay=20, tries=9, logger=LOG)
    def exec(self, command, stderr=True, stdin=False,
             stdout=True, tty=False, _preload_content=True,
             _request_timeout=120, **kwargs):
        """
        :param stderr: redirect stderr. Defaults to True
        :param stdin: redirect stdin. Defaults to False
        :param stdout: redirect stdout. Defaults to True
        :param tty: allocate tty. Defaults to False
        :param _preload_content: read/decode response data. Defaults to True
        # https://github.com/kubernetes-client/python/issues/10850
        :param _request_timeout: timeout for request. Defaults to 120
        :param container: container where command will be executed. optional
        """
        resp = stream(self._manager.api.connect_get_namespaced_pod_exec,
                      self.name, self.namespace,
                      command=command,
                      stderr=stderr, stdin=stdin,
                      stdout=stdout, tty=tty,
                      _preload_content=_preload_content,
                      _request_timeout=_request_timeout, **kwargs)
        return resp

    def wait_phase(self, phases, log_path=settings.ARTIFACTS_DIR,
                   timeout=500, interval=3):
        if isinstance(phases, str):
            phases = [phases]

        try:
            helpers.wait(lambda: self.read().status.phase in phases,
                         timeout=timeout, interval=interval,
                         timeout_msg='Timeout waiting, pod {0} phase is not '
                                     'in "{1}"'.format(self.name, phases))
        except Exception:
            pod_describe = str(self.read())
            pod_name = self.read().metadata.name
            timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
            logfile = '{path}pod-describe-{pod_name}-{timestamp}.log'.format(
                path=log_path,
                pod_name=pod_name,
                timestamp=timestamp
            )
            with open(logfile, 'w') as f:
                f.write(pod_describe)
            LOG.error(pod_describe)
            raise
        return self

    def wait_ready(self, log_path=settings.ARTIFACTS_DIR,
                   timeout=360, interval=3):
        try:
            helpers.wait(lambda: False not in [
                x.ready for x in self.read().status.container_statuses],
                         timeout=timeout, interval=interval,
                         timeout_msg='Timeout waiting, pod {0} is '
                                     'not Ready yet'.format(self.name))
        except Exception:
            pod_describe = str(self.read())
            pod_name = self.read().metadata.name
            timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
            logfile = '{path}pod-describe-{pod_name}-{timestamp}.log'.format(
                path=log_path,
                pod_name=pod_name,
                timestamp=timestamp
            )
            with open(logfile, 'w') as f:
                f.write(pod_describe)
            LOG.error(pod_describe)
            raise
        return self

    def wait_test(self, filepath, log_path=settings.ARTIFACTS_DIR,
                  timeout=600, interval=3):
        """Method that waits till test finishes.

        :param filepath: file that indicates test has finished
        :param timeout: time to wait before fail
        :param interval: interval between checks
        """
        try:
            helpers.wait(lambda: self.check_file_exists(filepath),
                         timeout=timeout, interval=interval,
                         timeout_msg='Timeout waiting, pod {0} does not have '
                                     'file "{1}"'.format(self.name, filepath))
            res = self.exec(
                ['/bin/sh', '-c', 'ls -la {0} || true'.format(
                     filepath)]).strip()
            LOG.info("In the pod {0}, expected the file {1}, found:\n{2}"
                     .format(self.name, filepath, res))
        finally:
            pod_name = self.read().metadata.name
            logs = self._manager.api.read_namespaced_pod_log(
                name=pod_name,
                namespace=self.read().metadata.namespace
            )
            timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
            logfile = '{path}pod-{pod_name}-{timestamp}.log'.format(
                path=log_path,
                pod_name=pod_name,
                timestamp=timestamp
            )
            with open(logfile, 'w') as f:
                f.write(logs)
            LOG.info(logs)
        return self

    def get_owner_reference(self):
        if self.exists():
            pod_meta = self.read().to_dict()['metadata']
        else:
            raise Exception("Cannot find pod {}".format(self.name))
        if pod_meta['owner_references']:
            kind = pod_meta['owner_references'][0]['kind']
            name = pod_meta['owner_references'][0]['name']
            return "{kind}/{name}".format(kind=kind, name=name)
        else:
            return "Node/no_owner"

    @property
    def data(self):
        """Returns dict of k8s object

        Data contains keys like api_version, kind, metadata,
        spec, status or items
        """
        return self.read().to_dict()


class K8sPodManager(K8sBaseManager):
    resource_class = K8sPod

    @property
    def api(self):
        return self._cluster.api_core

    def _list(self, namespace, **kwargs):
        return self.api.list_namespaced_pod(namespace, **kwargs)

    def _list_all(self, **kwargs):
        return self.api.list_pod_for_all_namespaces(**kwargs)

    def list_starts_with(self, pattern, namespace=None,
                         sort_by_len=False, **kwargs):
        """List pods that have pattern at the beginning of their names.
           Pods are fetched from all namespaces (by default) or
           from specified namespace. Method returns all pods with the same
           name prefix. sort_by_len sorts resulting list by pods name length.
        """
        if not namespace:
            pods = [pod for pod in self.list_all(**kwargs)
                    if pod.name.startswith(pattern)]
        else:
            pods = [pod for pod in self.list(namespace=namespace, **kwargs)
                    if pod.name.startswith(pattern)]
        if sort_by_len:
            # sort pods by name length
            pods.sort(key=lambda x: len(x.name))
        return pods

    def get_pods_statuses(self, pods):
        pods_info = {}
        for pod in pods:
            pod_name = pod['metadata']['name']
            pod_phase = pod['status']['phase']
            pod_message = pod['status']['message']
            if 'container_statuses' not in pod['status'] or \
                    not pod['status']['container_statuses']:
                pod_ready = {"no_containers_found": False}
            else:
                pod_ready = {x['name']: x['ready']
                             for x in pod['status']['container_statuses']}
            pods_info[pod_name] = (pod_phase, pod_message, pod_ready)
        return pods_info

    def check_pods_statuses(self, target_namespaces=None,
                            phases=('Running', 'Succeeded'),
                            excluded_pods=[], pods_prefix=''):

        failed_pods = self.list_not_in_phases(
            target_namespaces=target_namespaces,
            phases=phases)

        ready_text = " and/or Ready=False for " \
                     "some containers"

        if pods_prefix:
            failed_pods = \
                {k: v for k, v in failed_pods.items()
                 if k.startswith(pods_prefix)}

        for excl_pod in excluded_pods:
            failed_pods = \
                {k: v for k, v in failed_pods.items()
                 if not k.startswith(excl_pod)}

        if len(failed_pods) > 0:
            LOG.warning("Next pods are not in one of {} phase {}: \n{}".format(
                phases, ready_text,
                yaml.dump(failed_pods, default_flow_style=False)))
            return False
        return True

    def list_not_in_phases(self, target_namespaces=None,
                           phases=('Running', 'Succeeded'),
                           check_jobs=False,
                           excluded_nodes=None):
        pods = self.list_raw().to_dict()['items']
        if target_namespaces:
            if isinstance(target_namespaces, str):
                target_namespaces = [target_namespaces]
            pods = [pod for pod in pods
                    if pod['metadata']['namespace'] in target_namespaces]
        assert len(pods) > 0, "List of pods is empty"

        if not check_jobs:
            LOG.debug("Filter out Jobs from pod list")
            pods = \
                [x for x in pods if not (
                        x['metadata']['owner_references'] and
                        x['metadata']['owner_references'][0]['kind'] == 'Job')]
            LOG.debug(pods)
        if excluded_nodes:
            LOG.debug("Return pods from all, except {} nodes".format(
                excluded_nodes))
            # if None is in excluded_nodes list
            # pods w/o scheduled node will be filtered out
            pods = [x for x in pods
                    if x['spec']['node_name'] not in excluded_nodes]
            LOG.debug(pods)

        pods_info = self.get_pods_statuses(pods)
        failed_pods = {pod: (phase[0], phase[2])
                       for pod, phase in pods_info.items()
                       if pods_info[pod][0] not in phases or
                       (False in pods_info[pod][2].values() and
                        pods_info[pod][0] not in 'Succeeded')}
        return failed_pods

    def status(self, name, namespace):
        """Get status of the specified pod in the namespace"""
        return self.api.read_namespaced_pod_status(name, namespace)

    def present(self, name, namespace):
        """Check that specified pod name exists in the namespace"""
        return any([pod.name for pod in self.list(namespace=namespace)
                    if pod.name == name])
