from datetime import datetime, timedelta
from contextlib import contextmanager
from functools import partial
import pytz
import os

from si_tests.deployments.utils import commons, kubectl_utils, file_utils
from si_tests.deployments.utils.namespace import NAMESPACE
from si_tests.utils import waiters

BACKUP_JOB_NAME = "mariadb-phy-backup"
RESTORE_JOB_NAME = "mariadb-phy-restore"
OS_NAMESPACE = NAMESPACE.openstack


def assert_clean_env(os_manager):
    def assert_no_job(name, log_str):
        commons.LOG.info("Assert there is no %s", log_str)
        assert len(os_manager.api.jobs.list(
            namespace=OS_NAMESPACE, name_prefix=name)) == 0

    checks = [
        (BACKUP_JOB_NAME, "backup job"),
        (RESTORE_JOB_NAME, "restore job"),
    ]

    for check in checks:
        assert_no_job(*check)

    commons.LOG.info("Assert there is created cronjob %s in suspended state",
                     BACKUP_JOB_NAME)

    cronjob = os_manager.api.cronjobs.get(BACKUP_JOB_NAME,
                                          namespace=OS_NAMESPACE)
    assert cronjob.read().spec.suspend is True


def wait_for_new_job(os_manager, job_name, filter_func=None, filter_args=None):
    filter_args = filter_args or []
    jobs = os_manager.api.jobs.list(
        namespace=OS_NAMESPACE, name_prefix=job_name)
    if filter_func:
        jobs = list(filter(lambda job: filter_func(job, *filter_args), jobs))
    assert len(jobs) == 1
    return jobs[0]


def wait_job_complete(run_job):
    job = run_job.read()
    if job.status.failed:
        raise ValueError("Job {} failed to complete".format(run_job.name))
    assert job.status.succeeded == 1


def wait_for_job(os_manager, job_name, filter_func=None, filter_args=None,
                 complete_timeout=600):
    commons.LOG.info("Wait for %s job to trigger", job_name)

    waiter = partial(waiters.wait_pass, expected=AssertionError, interval=30)
    run_job = waiter(
        wait_for_new_job, timeout=240,
        predicate_args=(os_manager, job_name, filter_func, filter_args))

    commons.LOG.info("Wait for %s job to succeed", job_name)
    waiter(wait_job_complete, timeout=complete_timeout,
           predicate_args=(run_job,))

    commons.LOG.info("%s: \n%s\n", run_job.name, run_job.read().status)

    return run_job


def get_backup_dict(**backup):
    return {"spec": {"features": {"database": {"backup": backup}}}}


@contextmanager
def generate_backups(os_client, osdpl, os_manager):
    def get_time_for_cron():
        schedule = datetime.now() + timedelta(minutes=2)
        return schedule.astimezone(pytz.utc).strftime(
            "%M %H %d %m *")

    def filter_known(job, known):
        return job.name not in known

    def make_backup(known):
        osdpl.patch(get_backup_dict(backup_type="incremental", enabled=True,
                                    schedule_time=get_time_for_cron()))
        return wait_for_job(
            os_manager, BACKUP_JOB_NAME, filter_func=filter_known,
            filter_args=(known,))

    with os_client.provide_vm("test-vm-0") as vm0_id:
        commons.LOG.info("Perform base backup")
        run_job = make_backup([])

        with os_client.provide_vm("test-vm-1") as vm1_id:
            commons.LOG.info("Perform incremental backup")
            make_backup([run_job.name])
    yield vm0_id, vm1_id


@contextmanager
def spawn_backup_pod(os_manager):
    commons.LOG.info("Spawn pod to get backup name")
    kubectl = kubectl_utils.Kubectl()
    basepath = os.path.dirname(os.path.abspath(__file__))
    pod_yaml = file_utils.join(
        basepath, "../yamls/mariadb_check_pod.yaml")
    kubectl.apply(pod_yaml, cwd=".")
    try:
        def _wait_pod():
            pod = os_manager.api.pods.get(
                "si-tests-helper", namespace=OS_NAMESPACE)
            assert pod.read().status.phase == "Running"
            return pod

        commons.LOG.info("Wait for si-tests-helper pod to become Running")
        check_pod = waiters.wait_pass(
            _wait_pod, expected=Exception, timeout=120)
        yield check_pod
    finally:
        kubectl.delete("", "-f {}".format(pod_yaml), OS_NAMESPACE, cwd=".")


def get_backups_names(os_manager):
    def check_datetime(name):
        try:
            datetime.strptime(name, "%Y-%m-%d_%H-%M-%S")
        except ValueError as e:
            commons.LOG.exception("Backup name %s is not valid", name)
            raise e

    with spawn_backup_pod(os_manager) as check_pod:
        def get_files(path):
            return check_pod.exec(
                ["ls", os.path.join("/var/backup/", path)]).strip().split("\n")

        commons.LOG.info("Get base backup name")
        backup_name = get_files("base")[0]

        check_datetime(backup_name)

        commons.LOG.info("Found backup %s", backup_name)

        back_files = get_files(
            "base/{}".format(backup_name))

        back_files = list(map(lambda x: x.strip(), back_files))

        assert "backup.stream.gz" in back_files
        assert "backup.successful" in back_files

        commons.LOG.info("Get incremental backups")
        incr_backups_folders = get_files("incr")

        map(check_datetime, incr_backups_folders)

        assert backup_name in incr_backups_folders

        incr_backups = get_files("incr/{}".format(
            backup_name))

        commons.LOG.info("Incremental backups are %s", incr_backups)

        for incr_backup in incr_backups:
            check_datetime(incr_backup)
            files = get_files("incr/{}/{}".format(
                backup_name, incr_backup))
            assert "backup.stream.gz" in files
            assert "backup.successful" in files

    return backup_name, incr_backups[0]


def get_restore_dict(**kwargs):
    return {"spec": {"services": {"database": {"mariadb": {
        "values": kwargs}}}}}


def restore_backups(os_manager, osdpl, os_client, base_backup, incr_backup,
                    vm_ids):
    def filter_timestamp(job, known):
        return job.read().status.start_time not in known

    def restore(backup_name, vm_num, vm_ids, known_stamps):
        osdpl.patch(get_restore_dict(
            conf={"phy_restore": {"backup_name": backup_name}},
            manifests={"job_mariadb_phy_restore": True}))

        run_job = wait_for_job(
            os_manager, RESTORE_JOB_NAME, filter_func=filter_timestamp,
            filter_args=(known_stamps,), complete_timeout=1200)

        commons.LOG.info("Assert we have %s vm", vm_num)
        vms = os_client.server.list([])
        assert len(vms) == vm_num
        for vm in vms:
            assert vm["ID"] in vm_ids
        return run_job.read().status.start_time

    commons.LOG.info("Apply base restore")
    ran_timestamp = restore(base_backup, 1, vm_ids=vm_ids[:1], known_stamps=[])

    commons.LOG.info("Turn off the restore")
    osdpl.patch(get_restore_dict(
        manifests={"job_mariadb_phy_restore": False}))

    def wait_restore_deleted():
        assert len(os_manager.api.jobs.list(
            namespace=OS_NAMESPACE, name_prefix=RESTORE_JOB_NAME)) == 0

    commons.LOG.info("Wait for first restore job to be deleted")
    waiters.wait_pass(wait_restore_deleted, expected=AssertionError,
                      interval=30, timeout=240)

    commons.LOG.info("Apply incremental restore")
    restore("{}/{}".format(base_backup, incr_backup), 2, vm_ids=vm_ids,
            known_stamps=[ran_timestamp])


def mariadb_backup_restore(os_manager, openstack_client_manager):
    """ Test scenario:
        1. Create vm.
        2. Create base backup.
        3. Create second vm.
        4. Create incremental backup.
        5. Remove both vms.
        6. Spawn mariadb_check_pod to get backups names.
        7. Apply restore of the base backup.
        8. Check that we have 1 vm.
        9. Turn off restore.
        9. Apply incremental backup.
        10. Check that we have 2 vms.
    """
    assert_clean_env(os_manager)
    commons.LOG.info("Get osdpl")
    osdpl = os_manager.get_osdpl_deployment()

    try:
        with generate_backups(openstack_client_manager,
                              osdpl, os_manager) as vm_ids:
            backup_name, incr_backup = get_backups_names(os_manager)

            restore_backups(os_manager, osdpl, openstack_client_manager,
                            backup_name, incr_backup, vm_ids)
    finally:
        commons.LOG.info("Clean up vms created during restore")
        vms = ["test-vm-0", "test-vm-1"]
        for vm_name in vms:
            openstack_client_manager.server.delete([vm_name])
