import os
import re
from tempfile import NamedTemporaryFile
from time import sleep
from si_tests.utils.waiters import RunLimit

import yaml
from exec_helpers import CalledProcessError
from packaging import version
from si_tests import logger
from si_tests import settings
from si_tests.managers.bootstrap_manager import BootstrapManager
from si_tests.managers.kaas_manager import Manager as KaasManager

LOG = logger.logger


class KubeadmManager(BootstrapManager):
    def __init__(self):
        self.helm_globals = {}

    def get_latest_chart_version(self, remote, chartName):
        """
        helm unable to parse correctly latest chartVersion:
        $ helm search kaas-ipam --devel --versions |grep '2.8.0'
        kaas-bm/kaas-ipam       2.8.0         IPAM service
        kaas-bm/kaas-ipam       2.8.0-        IPAM service
        kaas-bm/kaas-ipam       2.8.0-1       IPAM service << the latest one
        But
        $ helm install kaas-bm/kaas-ipam --devel
        $ helm list
        NAME            REVISION        UPDATED                         STATUS          CHART           APP VERSION     NAMESPACE # noqa
        morbid-lion     1               Tue Jun 30 13:02:04 2020        DEPLOYED        kaas-ipam-2.8.0                 default # noqa
        """
        vers = remote.check_call("helm search {0} --devel"
                                 " --versions --output yaml".format(chartName))
        vers = [y['Version'] for y in vers.stdout_yaml]
        # Dont use clear version here, since we have non-sem-ver
        # -mcp- versions ;(
        vers.sort(key=lambda r: version.LegacyVersion(r))
        LOG.info(f'Guessed {chartName} chart version {vers[-1]}')
        return vers[-1]

    def helm_install(self, remote, args):
        if self.helm_globals and isinstance(self.helm_globals, dict):
            args += ' --set '
            args += ','.join(['global.{}={}'.format(k, v) for k, v in
                              self.helm_globals.items()])
        remote.check_call('helm install {}'.format(args), verbose=True)

    def wait_kubectl_status(self, remote, resource, json_path,
                            required_value='',
                            selector=None, timeout=600, interval=5,
                            required_value_re=''):
        if required_value:
            msg = required_value
        elif required_value_re:
            msg = "regexep for <{}>".format(required_value_re)
        else:
            raise ValueError("required_value or required_value_re must be "
                             "passed")
        if selector:
            LOG.info("Wait until resource's '{}' (selector '{}') value '{}'"
                     " becomes '{}'".format(resource, selector, json_path,
                                            msg))
        else:
            LOG.info("Wait until resource's '{}' value '{}' becomes '{}'"
                     .format(resource, json_path, msg))

        cmds = []
        if selector:
            if isinstance(selector, list):
                for item in selector:
                    cmds.append("kubectl get {} --selector {} -o jsonpath='{}'"
                                .format(resource, item, json_path))
            else:
                cmds.append("kubectl get {} --selector {} -o jsonpath='{}'"
                            .format(resource, selector, json_path))
        else:
            cmds.append("kubectl get {} -o jsonpath='{}'"
                        .format(resource, json_path))

        for cmd in cmds:
            LOG.info("Will use '{}' to query status, interval: {}, timeout: {}"
                     .format(cmd, interval, timeout))

        LOG.info("Expected status: {}".format(msg))

        last_values = {}
        with RunLimit(timeout, 'Timeout exceeded'):
            while True:
                for cmd in cmds:
                    res = remote.execute(cmd)

                    if len(res.stdout):
                        value = res.stdout[0].decode()
                    else:
                        value = None

                    last_value = last_values.setdefault(cmd, None)
                    if value != last_value:
                        LOG.info("Status has changed: '{}' --> '{}' ({})"
                                 .format(last_value, value, cmd))
                        last_values[cmd] = value

                    if required_value_re and value:
                        if re.match(re.compile(required_value_re),
                                    value.lower()):
                            LOG.info("Reached expected status.")
                            return True
                    else:
                        if value == required_value:
                            LOG.info("Reached expected status.")
                            return True

                sleep(interval)

    def step_004_deploy_kaas_bm_single(self, openstack_client):
        remote = self.remote_seed(openstack_client)
        seed_ip = self.get_seed_ip(openstack_client)
        seed_user = settings.SEED_SSH_LOGIN
        seed_key = settings.SEED_SSH_PRIV_KEY_FILE
        LOG.info('Upload kaas-bm-env-key to seed_node')
        os.system("rsync -e \"ssh -i {0} -o StrictHostKeyChecking=no\" "
                  "-azqP {0} {1}@{2}:"
                  "/home/ubuntu/kaas-bm-env-key".
                  format(seed_key, seed_user,
                         seed_ip))
        LOG.info('Base setup, render and create VMs using virsh')

        self.ansible_init(openstack_client)
        ansible_vars = {
            "virt_lab": "yes",
            "deployment_modes": ["virtual", "single_test"]
        }

        self.ansible_run(remote, "./setup-playbook.yml",
                         inventory="./inventory_virt_deploy.yaml",
                         extra_vars=ansible_vars)

        LOG.info('Bootstrap KaaS-BM single node')
        remote.reconnect()

        # Insane 2h timeout related: PRODX-3539
        ret = self.ansible_run(remote, "./setup-kaas-bm-single.yml",
                               inventory="./inventory_virt_deploy.yaml",
                               raise_on_err=False,
                               timeout=60 * 60 * 2)

        LOG.info('Try fetch logs')
        if remote.isfile('./log_bootstrap.log'):
            remote.check_call('tar zcfv logs.tar.gz ./log_bootstrap.log')
            remote.download("./logs.tar.gz",
                            "{0}/management_logs.tar.gz"
                            .format(settings.ARTIFACTS_DIR))
            LOG.info("Remote logs placed to {0}"
                     "/management_logs.tar.gz".format(settings.ARTIFACTS_DIR))

        if ret.exit_code == 0:
            LOG.info("Mgm cluster has been deployed successfuly")
        else:
            LOG.error("Mgm cluster has not been deployed successfuly")
            raise CalledProcessError(result=ret, expected=(0,))

    def step_005_test_bm_helm_charts(self, openstack_client):
        # TODO: refactor to kaasapi client completly
        # extend with ssh-to-provisioned-node check
        """
         1) Apply all resources
         8) Broke test subnet ('kaas-mgmt' - hardcoded in ansible)

        :param openstack_client:
        :return:
        """
        testNSname = 'test1-ns'
        remote = self.remote_seed(openstack_client)
        kaas_core_remote_dir = settings.KAAS_CORE_REMOTE_PATH

        kubeconfig_old = self.get_remote_file_yaml(openstack_client,
                                                   ".kube/config")
        seed_ip = self.get_seed_ip(openstack_client, fip='seed_floating_ip')
        # Get kubeconfig, with fixed server addr to external node IP
        with NamedTemporaryFile(mode="w+") as f:
            f.write(yaml.dump(kubeconfig_old))
            f.flush()
            kubeconfig = self.expose_kubeconfig(kubeconfig_old, seed_ip)

        kconffile = "{0}/management_kubeconfig".format(settings.ARTIFACTS_DIR)
        with open(kconffile, "w") as f:
            f.write(yaml.dump(kubeconfig))
        kaas_manager = KaasManager(kubeconfig=kconffile)

        LOG.info("Get/set helm globals")
        if settings.KAAS_BM_HELM_GLOBALS_DOCKER_BASE_URL:
            self.helm_globals['dockerBaseUrl'] = \
                settings.KAAS_BM_HELM_GLOBALS_DOCKER_BASE_URL
        if settings.KAAS_BM_HELM_GLOBALS_HELM_BASE_URL:
            self.helm_globals['helmBaseUrl'] = \
                settings.KAAS_BM_HELM_GLOBALS_HELM_BASE_URL
        if settings.KAAS_BM_HELM_GLOBALS_UBUNTU_CODENAME:
            self.helm_globals['lsb_codename'] = \
                settings.KAAS_BM_HELM_GLOBALS_UBUNTU_CODENAME
        LOG.info('helm_globals: %s', self.helm_globals)

        LOG.info("Add kaas-bm helm repo")
        repo_url = settings.KAAS_BM_CHARTS_REPO
        remote.check_call("helm repo add kaas-bm {}".format(repo_url),
                          verbose=True)
        remote.check_call("helm repo update", verbose=True)

        LOG.info("Install kaas-public-api")
        self.helm_install(remote, '{}/charts/kaas-public-api'.format(
            kaas_core_remote_dir))

        LOG.info("Add KaaSRelease objects")
        remote.check_call(
            "find {}/bootstrap/releases/kaas -name '*.yaml' "
            "-exec kubectl apply -f {{}} \\;".format(kaas_core_remote_dir))

        LOG.info("Add ClusterRelease objects")
        remote.check_call(
            "find {}/bootstrap/releases/cluster -name '*.yaml' "
            "-exec kubectl apply -f {{}} \\;".format(kaas_core_remote_dir))

        LOG.info("Patch KaasRelease objects")
        remote.check_call(
            "kubectl get kaasrelease -o jsonpath='{$.items[*].metadata.name}' "
            "| tr \" \" \"\n\" | xargs -I {} "
            "kubectl patch kaasrelease {} --type=json --patch='[{"
            "\"op\": \"add\","
            "\"path\": \"/metadata/labels\","
            "\"value\": {\"kaas.mirantis.com/active\": \"true\"}"
            "}]'", verbose=True)

        LOG.info("Install PersistentVolumes")
        remote.check_call('kubectl apply -f bootstrap/bm-operator-pv.yaml',
                          verbose=True)

        LOG.info("Install lcm-controller")
        self.helm_install(remote,
                          '{}/charts/lcm-controller '.format(
                              kaas_core_remote_dir))
        LOG.info('Scale lcm-controller to 1')
        remote.check_call(
            'kubectl scale deployment '
            'lcm-lcm-controller --replicas=1',
            verbose=True)

        LOG.info("Install metallb")
        self.helm_install(remote,
                          '{}/charts/metallb '
                          '-f bootstrap/metallb.values.yaml'.format(
                              kaas_core_remote_dir))

        # Install baremetal-public-api
        kbpapi_helm_opts = ''
        if settings.KAAS_BM_PUBLIC_API_CHART_URL:
            kbpapi_helm_opts += f' {settings.KAAS_BM_PUBLIC_API_CHART_URL}'
        else:
            kbpapi_helm_opts += ' kaas-bm/baremetal-public-api '
            if settings.KAAS_BM_PUBLIC_API_CHART_VERSION:
                kbpapi_helm_opts += ' --version {}'.format(
                    settings.KAAS_BM_PUBLIC_API_CHART_VERSION)
            else:
                v = self.get_latest_chart_version(
                    remote, 'kaas-bm/baremetal-public-api')
                kbpapi_helm_opts += f' --version {v}'
        LOG.info("Install baremetal-public-api chart with args '%s'",
                 kbpapi_helm_opts)
        self.helm_install(remote, kbpapi_helm_opts)

        LOG.info("Install kaas-bm provider")
        self.helm_install(remote,
                          '-f bootstrap/baremetal-provider.values.yaml '
                          '{}/charts/baremetal-provider'.format(
                              kaas_core_remote_dir))
        LOG.info('Scale bm-provider to 1')
        remote.check_call(
            'kubectl scale deployment '
            'baremetal-provider --replicas=1',
            verbose=True)

        # Install baremetal-operator
        kbo_helm_opts = ' -f bootstrap/baremetal-operator.values.yaml '
        if settings.KAAS_BM_OPERATOR_CHART_URL:
            kbo_helm_opts += ' {}'.format(settings.KAAS_BM_OPERATOR_CHART_URL)
        else:
            kbo_helm_opts += ' kaas-bm/baremetal-operator '
            if settings.KAAS_BM_OPERATOR_CHART_VERSION:
                kbo_helm_opts += ' --version {}'.format(
                    settings.KAAS_BM_OPERATOR_CHART_VERSION)
            else:
                v = self.get_latest_chart_version(remote,
                                                  'kaas-bm/baremetal-operator')
                kbo_helm_opts += f' --version {v}'
        LOG.info("Install baremetal-operator chart with args '{}'".format(
            kbo_helm_opts))
        self.helm_install(remote, kbo_helm_opts)

        # Install kaas-ipam
        kbipam_helm_opts = ' -f bootstrap/kaas-ipam.values.yaml '
        if settings.KAAS_BM_IPAM_CHART_URL:
            kbipam_helm_opts += ' {}'.format(settings.KAAS_BM_IPAM_CHART_URL)
        else:
            kbipam_helm_opts += ' kaas-bm/kaas-ipam '
            if settings.KAAS_BM_IPAM_CHART_VERSION:
                kbipam_helm_opts += ' --version {}'.format(
                    settings.KAAS_BM_IPAM_CHART_VERSION)
            else:
                v = self.get_latest_chart_version(remote,
                                                  'kaas-bm/kaas-ipam')
                kbipam_helm_opts += f' --version {v}'
        LOG.info("Install kaas-ipam chart with args '{}'".format(
            kbipam_helm_opts))
        self.helm_install(remote, kbipam_helm_opts)

        LOG.info("Wait for ironic warm-up")
        pod_status = self.wait_kubectl_status(
            remote, resource='pod',
            # FIXME: when CR kaas-bm/metal3-helm-chart:65017 is merged
            #  remove old selector 'app.kubernetes.io/name==kaas-bm-ironic'
            #  from the list below.
            selector=['app.kubernetes.io/name==kaas-bm-ironic',
                      'app.kubernetes.io/name==ironic'],
            json_path='{$.items[0].status.phase}',
            required_value='Running',
            timeout=1200)
        if not pod_status:
            raise Exception("Failed to get pod running in time")

        if settings.KAAS_BM_IRONIC_OPERATOR_DOCKER_IMAGE:
            LOG.info("Wait for ironic-operator warm-up")
            pod_status = self.wait_kubectl_status(
                remote, resource='pod',
                selector='app.kubernetes.io/name==ironic-operator',
                json_path='{$.items[0].status.phase}',
                required_value='Running',
                timeout=300)
            if not pod_status:
                raise Exception("Failed to get pod running in time")

        LOG.info("install default baremetalhostprofiles")
        remote.check_call("kubectl apply -f bootstrap/dev/templates/bm/"
                          "baremetalhostprofiles.yaml.template")

        kaasreleases = kaas_manager.get_kaasrelease_names()
        latest_kaasrelease = str(max([version.parse(x) for x in kaasreleases]))
        kaas_release = kaas_manager.get_kaasrelease(latest_kaasrelease)
        kaas_release_name = kaas_release.data['metadata']['name']
        clusterRelease = kaas_release.data['spec']['clusterRelease']

        ns = kaas_manager.create_namespace(testNSname)

        # copy public key, to be passed into cluster
        mgmtKeyFile = "{0}/" \
                      "management_id_rsa_pub".format(settings.ARTIFACTS_DIR)
        public_key_name = 'common-public-key'
        if not remote.download(".ssh/id_rsa.pub",
                               mgmtKeyFile):
            raise Exception("Unable to download default system ssh public key")
        with open(mgmtKeyFile) as pub_key:
            pub_key_content = pub_key.read()
        ns.create_publickey(public_key_name, pub_key_content)

        # TODO those should be refactored to kaas_manager.ns.create_cluster
        LOG.info("Create dummy management cluster")
        mgmt_cluster_body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Cluster",
            "metadata": {
                "name": 'test1',  # hard-coded in ansible
                "annotations": {
                    "kaas.mirantis.com/lcm": "true"
                },
                "labels": {
                    # hardocded on ansible part
                    "kaas.mirantis.com/region": 'region-one',
                    "kaas.mirantis.com/provider": 'baremetal'
                }
            },
            "spec": {
                "clusterNetwork": {
                    "services": {
                        "cidrBlocks": [
                            '10.233.0.0/18'
                        ]
                    },
                    "pods": {
                        "cidrBlocks": [
                            '10.233.64.0/18'
                        ]
                    },
                    "serviceDomain": "cluster.local"
                },
                "providerSpec": {
                    "value": {
                        "apiVersion": kaas_manager.get_bmApiversion(
                            latest_kaasrelease),
                        "kind": "BaremetalClusterProviderSpec",
                        "dnsNameservers": [
                            '172.18.224.6'
                        ],
                        "nodeCidr": '10.10.10.0/24',
                        "release": clusterRelease,
                        "publicKeys": [
                            {"name": public_key_name}
                        ],
                        "dedicatedControlPlane": False,
                        "loadBalancerHost": "10.0.0.61",
                        "kaas": {
                            "management": {
                                "enabled": True},
                            "release": kaas_release_name,
                        }
                    }
                }
            }
        }
        kaas_manager.api.kaas_clusters.create(namespace=testNSname,
                                              body=mgmt_cluster_body)

        LOG.info("Create dummy regional cluster")
        regional_cluster_body = {
            "apiVersion": "cluster.k8s.io/v1alpha1",
            "kind": "Cluster",
            "metadata": {
                "name": 'test1',  # hard-coded in ansible
                "annotations": {
                    "kaas.mirantis.com/lcm": "true"
                },
                "labels": {
                    # hardocded on ansible part
                    "kaas.mirantis.com/region": 'region-one',
                    "kaas.mirantis.com/provider": 'baremetal'
                }
            },
            "spec": {
                "clusterNetwork": {
                    "services": {
                        "cidrBlocks": [
                            '10.233.0.0/18'
                        ]
                    },
                    "pods": {
                        "cidrBlocks": [
                            '10.233.64.0/18'
                        ]
                    },
                    "serviceDomain": "cluster.local"
                },
                "providerSpec": {
                    "value": {
                        "apiVersion": kaas_manager.get_bmApiversion(
                            latest_kaasrelease),
                        "kind": "BaremetalClusterProviderSpec",
                        "dnsNameservers": [
                            '172.18.224.6'
                        ],
                        "nodeCidr": '10.10.10.0/24',
                        "release": clusterRelease,
                        "publicKeys": [
                            {"name": public_key_name}
                        ],
                        "dedicatedControlPlane": False,
                        "loadBalancerHost": "10.0.0.61",
                        "kaas": {
                            "release": kaas_release_name,
                            "regional": [
                                {"helmReleases": [
                                    {"values": {
                                        "region": "region-one"},
                                        "name": "baremetal-operator"},
                                    {"values": {
                                        "config": {"region": "region-one"}},
                                        "name": "baremetal-provider"},
                                    {"values": {"region": "region-one"},
                                     "name": "kaas-ipam"}
                                ],
                                    "provider": "baremetal"
                                }]
                        }
                    }
                }
            }
        }
        kaas_manager.api.kaas_clusters.create(namespace='default',
                                              body=regional_cluster_body)

        LOG.info("Import IPAM L2Templates objects")
        remote.check_call('kubectl apply -R -f bootstrap/ipam/l2template.d',
                          verbose=True)

        LOG.info("Import ALIVE IPAM subnet objects")
        remote.check_call('kubectl apply -R -f bootstrap/ipam/subnet.d',
                          verbose=True)

        LOG.info("Wait for `subnet kaas-mgmt` active")
        if not self.wait_kubectl_status(
                remote, resource='subnet kaas-mgmt',
                json_path='{$.status.statusMessage}',
                required_value_re='(?i)^OK',
                timeout=120):
            raise Exception("Wait for `subnet kaas-mgmt` active - FAILED")

        LOG.info("Broke `subnet kaas-mgmt`")
        remote.check_call('kubectl patch subnet kaas-mgmt '
                          '--type=\'json\' -p=\'[{"op": "replace", '
                          '"path": "/spec/gateway", "value":"10.0.1.1"}]\'')

        if not self.wait_kubectl_status(
                remote, resource='subnet kaas-mgmt',
                json_path='{$.status.statusMessage}',
                required_value_re='(?i)^ERR',
                timeout=120):
            raise Exception("Wait for broke `subnet kaas-mgmt` - FAILED")

        LOG.info("Import IPAM IP objects")
        remote.check_call('kubectl apply -R -f bootstrap/ipam/ip.d',
                          verbose=True)

        LOG.info("Creating baremetal hosts")
        # quick fix http svc path
        remote.execute('cd bootstrap/hw/mgmt.d/;'
                       'export svc_ip=$(kubectl get svc httpd-http -o '
                       'jsonpath="{.status.loadBalancer.ingress[0].ip}");'
                       'for i in *;'
                       'do sed -i "s/SED_ME_IN_TEST_HTTPD_SVC/$svc_ip/g" '
                       '"$i"; done')
        remote.check_call("kubectl apply -R -f bootstrap/hw/mgmt.d/",
                          verbose=True)
        remote.check_call("kubectl apply -R -f"
                          " bootstrap/machine/mgmt.d/",
                          verbose=True)
        # # Here we should have ipamhosts objects, but in failed state
        # # It may appear only after introspection
        LOG.info("Wait for broken ipamhosts mgmt-0")
        if not self.wait_kubectl_status(
                remote, resource=f'-n {testNSname} ipamhosts mgmt-0',
                json_path='{$.status.ipAllocationResult}',
                required_value_re='.*failed.*',
                timeout=900,
                interval=30):
            raise Exception("Wait for broken `ipamhosts mgmt-0` - FAILED")

        LOG.info("Fix back `subnet kaas-mgmt`")
        remote.check_call('kubectl patch subnet kaas-mgmt '
                          '--type=\'json\' -p=\'[{"op": "replace", '
                          '"path": "/spec/gateway", "value":"10.0.0.1"}]\'')
        if not self.wait_kubectl_status(
                remote, resource=f'-n {testNSname} ipamhosts mgmt-0',
                json_path='{$.status.ipAllocationResult}',
                required_value='OK',
                timeout=900):
            raise Exception("Wait for Fix `ipamhosts mgmt-0` - FAILED")

        # Continue deployment check
        LOG.info('Wait for baremetal host provisioned')
        ns.wait_baremetalhosts_statuses(retries=60, interval=40)

        # Those part must be refactored to use 'bastion' logic, aka:
        # cluster = ns.get_cluster(name='test1')
        # cluster.bastion_ip = seed_ip
        # m1 = self.get_machines(machine_type='control',
        #                                   machine_status='Ready')[0]
        # m1.run_cmd('uname -a', verbose=True)
        # after implementation https://mirantis.jira.com/browse/TASA-3
        ipamhosts = [i.read().to_dict() for i in ns.get_ipam_hosts()]

        LOG.info("Wait for L2Template rendering on ipamhosts")
        for ipamhost in ipamhosts:
            ipamhost_name = ipamhost['metadata']['name']
            LOG.info("IpamHost to check: {}".format(ipamhost_name))
            if not self.wait_kubectl_status(
                    remote,
                    resource=f'-n {testNSname} ipamhost {ipamhost_name}',
                    json_path='{$.status.l2RenderResult}',
                    required_value='OK',
                    selector='',
                    timeout=900,
                    interval=30):
                raise Exception("Wait for L2Template rendering on"
                                " {} - FAILED".format(ipamhost_name))

        ips = [i['status']['osMetadataNetwork']['networks'][0]['ip_address']
               for i in ipamhosts]
        assert len(ips) == len(ipamhosts), \
            "Some of IpamHosts does not contain IPs for provisioning network"
        ssh_cmd = 'ssh -o ConnectTimeout=10 ' \
                  '-o StrictHostKeyChecking=no ' \
                  'mcc-user@{ip} "uname -a;df -h;' \
                  'uptime;' \
                  'curl -q --connect-timeout 5 --max-time 5 ' \
                  '--retry 5 --retry-delay 1 --retry-max-time 5 ' \
                  'http://mirror.mirantis.com -o /dev/null"'
        failed = []
        for ip in ips:
            retries = 5
            # Ssh should be correctly configured at seed node for that time.
            # Also, nodes might spin-up to a bit long time, so we need to wait
            # a little bit
            LOG.info('Check ssh access to: <<{}>>'.format(ip))
            r_sleep = 30
            for r in range(1, retries + 1):
                ret = remote.check_call(ssh_cmd.format(ip=ip), verbose=True,
                                        raise_on_err=False)
                if ret.exit_code == 0:
                    break
                elif ret.exit_code != 0 and r < retries:
                    LOG.info('Unable check ssh access to: <<{}>> attempt'
                             ' {}/{}/sleep:{}'.format(ip, r, retries, r_sleep))
                    sleep(r_sleep)
                    continue
                elif ret.exit_code != 0:
                    failed.append(ip)
        if failed:
            raise Exception(
                'Unable to check ssh access to freshly provisioned '
                'node addresses: {}'.format(','.join(failed)))
        LOG.info('All single tests finished')
