from si_tests import logger
import pytest

LOG = logger.logger


@pytest.fixture(scope='module')
def iam_api_client(keycloak_admin_client, iam_api):
    # We create "test" user to use it in scenarios
    # maybe later we will use ldap or openid user
    username = 'si-test'
    password = '123'
    keycloak_admin_client.user_create(username=username, password=password)
    # add admin role
    keycloak_admin_client.assign_roles(username=username,
                                       roles=["m:iam@admin"])
    client = iam_api.connect(username=username, password=password)
    yield client
    keycloak_admin_client.user_delete(username=username)


def test_iam_api_versions_and_token(iam_api_client):
    version = iam_api_client.get_version()
    assert version.ApiVersion == '1.0.0', \
        "get_version returned unknown version number: {}".format(version)
    token = iam_api_client.get_token
    assert token, \
        "Token is not retrieved"


def test_iam_api_list_users(iam_api_client):
    filter = {"Username": "iamadmin"}
    iamadmin = iam_api_client.list_users(filter=filter)[0]
    assert iamadmin.username == "iamadmin",\
        "iamadmin is not found: {}".format(iamadmin)


def test_iam_api_get_user(iam_api_client):
    test_user = iam_api_client.get_this_user()
    assert test_user.username == "si-test",\
        "si-test is not found: {}".format(test_user)
    test_user2 = iam_api_client.get_user('si-test')
    assert test_user2 == test_user,\
        "si-test is not found: {}".format(test_user2)


def test_iam_api_give_revoke_grants(iam_api_client):
    # iam
    # give grant/role
    username = iam_api_client.get_this_user().username
    role = 'reader'
    scope = 'm:kaas'
    iam_api_client.give_grants(username=username, role=role,
                               scope=scope)
    print(iam_api_client.list_grants(username=username))
    grants = [{'role': role, 'scope': scope}]
    result = iam_api_client.has_grants(username, grants)
    assert not result['missing_grants'], \
        "Some grants are not found is not found: {}".format(result)
    # revoke grant/role
    iam_api_client.revoke_grants(username=username, role=role,
                                 scope=scope)
    print(iam_api_client.list_grants(username=username))
    result = iam_api_client.has_grants(username, grants)
    assert not result['same'], \
        "Some grants are not found is not found: {}".format(result)


def test_iam_api_has_grants(iam_api_client):
    # iam
    # give grant/role
    username = iam_api_client.get_this_user().username
    iam_api_client.give_grants(username=username, role='reader',
                               scope='m:kaas')
    grants = [{'role': 'reader', 'scope': 'm:kaas'}]
    result = iam_api_client.has_grants(username, grants)
    # we expect that user CAN give grants to itself
    assert not result['missing_grants'], \
        "Grant lists are not the same. Result: {0}".format(result)
    iam_api_client.give_grants(username=username, role='writer',
                               scope='m:kaas')
    iam_api_client.give_grants(username=username, role='admin',
                               scope='m:os')
    grants = [{'role': 'reader1', 'scope': 'm:kaas'},
              {'role': 'viewer', 'scope': 'm:iam'}]
    result = iam_api_client.has_grants(username, grants)
    assert result['no_matches'] and \
        result['missing_grants'], \
        "Grant lists must have no matches, " \
        "but there are some. Result: {0}".format(result)
    grants = [{'role': 'reader', 'scope': 'm:kaas'},
              {'role': 'viewer', 'scope': 'm:iam'}]
    result = iam_api_client.has_grants(username, grants)
    assert not result['no_matches'] and \
        not result['same'] and \
        result['missing_grants'] and \
        result['extra_grants'], \
        "Something is wrong. Result: {0}".format(result)


def test_iam_api_give_my_grants(iam_api_client):
    print(iam_api_client.list_my_grants())


def test_iam_api_list_scopes(iam_api_client):
    scopes = [x.name for x in iam_api_client.list_scopes()]
    assert "m:kaas" in scopes, \
        "m:kaas scope is not found: {}".format(scopes)


def test_iam_api_list_roles_for_scope(iam_api_client):
    scope = "m:kaas"
    roles = [x.name for x in iam_api_client.list_roles_for_scope(scope=scope)]
    assert "reader" in roles, \
        "reader role is not found in scope {0}: {1}".format(scope, roles)


def test_iam_api_list_users_for_scopes(iam_api_client):
    scope = "m:kaas"
    users = [x.user.username for x in iam_api_client.list_users_for_scope(
        scope=scope
    )]
    assert "reader" in users, \
        "reader user is not found in scope {0}: {1}".format(scope, users)


def test_update_user(keycloak_admin_client, iam_api_client):
    username = iam_api_client.get_this_user().username
    # keycloak
    keycloak_admin_client.user_update(username=username,
                                      body={'firstName': 'updated'})
    test_user = iam_api_client.get_this_user()
    assert test_user.metadata['FirstName'] == 'updated', \
        "{0} was not updated: {1}".format(username, test_user)


def test_servicetoken(iam_api_client):
    service = 'kaas'
    alias = 'alias'
    print(iam_api_client.list_servicetoken())
    print(iam_api_client.list_servicetoken(all=False))
    grants = [{'role': "reader", 'scope': "m:{}".format(service)}]
    servicetoken = iam_api_client.create_servicetoken(alias, service, grants)
    st = iam_api_client.get_servicetoken(id=servicetoken.id)
    assert st.alias == alias, \
        "Servicetoken alias is not correct: {}".format(st)
    assert st.grants[0].role == grants[0]['role'] and \
        st.grants[0].scope == grants[0]['scope'], \
        "Servicetoken grants are not correct: {}".format(st)
    iam_api_client.delete_servicetoken(id=servicetoken.id)
    result = [x.alias for x in iam_api_client.list_servicetoken()]
    assert alias not in result, \
        "Servicetoken was not deleted properly."


def test_iam_api_create_delete_user(keycloak_admin_client, iam_api):
    # just to show, that we can create additional users for tests
    keycloak_admin_client.user_create(username="user2", password='password',
                                      email='user@user.com')
    user2 = iam_api.connect(username='user2', password='password')
    print(user2.get_token())
    version = user2.get_version()
    assert version.ApiVersion == '1.0.0', \
        "get_version returned unknown version number: {}".format(version)
    keycloak_admin_client.user_delete(username='user2')
