import time
import yaml

from kubernetes import client as kube_client
from si_tests import settings
from si_tests.utils import waiters
from si_tests import logger
from si_tests.clients.k8s import K8sCluster

LOG = logger.logger


def test_add_ceph_osd_node():
    """This test adds additional osd node
       to already created ceph cluster. The scale
       node wich will be added should be already existed in
       kubernetes cluster.

       1. Get information about ceph cluster using ceph tools pod.
       2. Define node for scaling according to SCALE_NODE_NAME_PATTERN
       3. Form data for patch miraceph CRD
       4. Patch CRD and wait for correct status, num osds etc.

       Parameters required for test execution:
         - KUBECONFIG
         - SCALE_NODE_NAME_PATTERN
         - ROOK_CEPH_NS
         - CEPH_DEVICE_ROLE
         - CEPH_LCM_MIRANTIS_NS
    """

    scale_node_name_pattern = settings.SCALE_NODE_NAME_PATTERN
    rook_ns = settings.ROOK_CEPH_NS
    ceph_device_role = settings.CEPH_DEVICE_ROLE
    ceph_lcm_mirantis_ns = settings.CEPH_LCM_MIRANTIS_NS
    # Init k8s client
    client = K8sCluster(kubeconfig=settings.KUBECONFIG_PATH)
    # https://github.com/kubernetes-client/python/issues/234
    delete_options = kube_client.V1DeleteOptions(
        api_version='v1',
        kind="DeleteOptions",
        propagation_policy="Background")

    # Getting pods for checking Ceph status
    ceph_tools_pod = client.pods.list(namespace=rook_ns,
                                      name_prefix='rook-ceph-tools')
    assert len(ceph_tools_pod) > 0, ("No pods found with prefix "
                                     "rook-ceph-tools in namespace {}".format(
                                         rook_ns))
    ceph_operator_pod = client.pods.list(namespace=rook_ns,
                                         name_prefix='rook-ceph-operator')
    assert len(ceph_operator_pod) > 0, ("No pods found with prefix "
                                        "rook-ceph-operator in "
                                        "namespace {}".format(rook_ns))
    ceph_tools_pod = ceph_tools_pod[0]
    ceph_operator_pod = ceph_operator_pod[0]

    cmd_status = ['/bin/sh', '-c', 'ceph -s -f json']
    cmd_osd_tree = ['/bin/sh', '-c', 'ceph osd tree -f json']

    # Determine node for scale (Maybe we need to parametrize pattern)
    scale_node = [node for node in
                  client.nodes.list_all()
                  if scale_node_name_pattern in node.name]
    assert len(scale_node) > 0, ("No nodes with name pattern {} "
                                 "found for scaling.".format(
                                     scale_node_name_pattern))
    scale_node = scale_node[0]
    scale_node_data = scale_node.read().to_dict()

    # Getting required information for adding node
    scale_node_ip = [ip.get('address') for ip in scale_node_data.get(
        'status').get('addresses') if 'InternalIP' in ip.get('type')][0]
    mem_cap = scale_node_data.get('status').get('capacity').get('memory')
    cpu_cap = int(scale_node_data.get('status').get('capacity').get('cpu'))
    if 'Ki' in mem_cap:
        mem_cap = round(int(mem_cap.split('Ki')[0]) / 1024 / 1024)
    ceph_node_roles = []
    scale_node_name = scale_node.name

    LOG.info("Next node will be used for scaling ceph osd nodes: {}".format(
        scale_node_name))

    # Hack for using correct devices
    rook_discover_pod = [
        pod for pod in client.pods.list(
            namespace=rook_ns,
            name_prefix='rook-discover') if scale_node_name in
        pod.read().to_dict().get('spec').get('node_name')][0]

    # Device for osd should not contain any file system so we will
    # skip all of the most expected fat|lvm|ext|iso. Also, skip
    # all ".da" devices, because there is a host system. These
    # skips not a standart, but for tests purposes in our cases is
    # more then enough. Only one risk is existed, if we miss some
    # file system from filtering, then this device will not be
    # used for ceph osd. Ceph operator do nothing with such devices
    rook_pod_cmd = [
        '/bin/sh',
        '-c',
        '''lsblk --nodeps --fs --noheading --output KNAME,FSTYPE |
        grep -vEi ".da|fat|lvm|ext|iso"''']
    rook_pod_all_disks_cmd = [
        '/bin/sh',
        '-c',
        'lsblk --nodeps --fs --output KNAME,FSTYPE']
    disks = rook_discover_pod.exec(rook_pod_cmd).split('\n')
    disks_list = [disk.strip() for disk in disks if disk]
    assert len(disks_list) > 0, ("No disks available for ceph osd. "
                                 "Looks like all of disks for node {0} "
                                 "are used: {1} {2}").format(
        scale_node_name, "\n", rook_discover_pod.exec(rook_pod_all_disks_cmd))

    LOG.info("Next disks in node {} will be used for ceph OSDs: {}".format(
        scale_node_name, disks_list))
    # Forming storage devices list
    storagedevices = []
    for disk in disks_list:
        ceph_device_size = rook_discover_pod.exec([
            '/bin/sh',
            '-c',
            "lsblk --nodeps /dev/{} --output SIZE --noheadings".format(disk)])
        if 'T' in ceph_device_size:
            device_size_gb = round(float(ceph_device_size.split('T')[0])*1024)
        elif 'G' in ceph_device_size:
            device_size_gb = round(float(ceph_device_size.split('G')[0]))
        else:
            LOG.error("Device size is not correct value, or too small. "
                      "Please, check manually. Current device size for disk "
                      "{} is: {}".format(disk, ceph_device_size))
        storagedevices.append({"name": disk,
                               "role": ceph_device_role,
                               "sizeGb": device_size_gb})
    assert len(storagedevices) > 0, ("No storage devices found for using "
                                     "for ceph osds")

    # Getting miracephs CRD
    miraceph_crd = [crd for crd in client.miracephs.list(
        namespace=ceph_lcm_mirantis_ns) if crd.name == 'cephcluster']
    assert miraceph_crd, ("CRD with name cephcluster is not found")
    miraceph_crd = miraceph_crd[0]
    # Getting client/cluster ceph networks
    client_net = miraceph_crd.data.get(
            'spec', {}).get('clientNet', '')
    cluster_net = miraceph_crd.data.get(
            'spec', {}).get('clusterNet', '')

    # Forming data for patching miracephs CRD
    patch_ceph_data = {"spec": {
        "clientNet": client_net,
        "clusterNet": cluster_net,
        "nodes": {
            scale_node_name: {
                "cores": cpu_cap,
                "crushPath": {},
                "ips": [
                    scale_node_ip
                ],
                "ramGb": mem_cap,
                "roles": ceph_node_roles,
                "storageDevices": storagedevices
            }
        }
    }}

    LOG.info("Cephcluster CRD will be patched with next data:{}{}".format(
        "\n", yaml.dump(patch_ceph_data)))
    # Check if HEALTH_OK before patching

    LOG.info("Check if HEALTH_OK before patching CRD")
    ceph_health_timeout = 2400
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))
    LOG.info("Ceph health is HEALTH_OK. Continue...")

    # Get osds num before patch and calculate num after
    num_osds_before_patch = int(
        yaml.safe_load(ceph_tools_pod.exec(
            cmd_status)).get('osdmap').get('osdmap').get('num_osds'))
    num_osds_after_patch = num_osds_before_patch + len(disks_list)

    LOG.info("Patching miraceph cephcluster CRD")
    # Patch CRD with new node
    miraceph_crd.patch(body=patch_ceph_data)
    # Need a few seconds after patching before removing pods
    time.sleep(5)

    # Delete operator pod and osd-prepare jobs (documentation)
    LOG.info("Restarting ceph-operator pod and osd-prepare jobs")
    ceph_operator_pod.delete()
    osd_prepare_jobs = client.jobs.list(
        namespace=rook_ns, name_prefix='rook-ceph-osd-prepare')
    for job in osd_prepare_jobs:
        client.api_batch.delete_namespaced_job(
            namespace=rook_ns, name=job.name, body=delete_options)

    # Wait for num osds is inceased depending on disks count
    LOG.info("Waiting for OSDs number is increased in ceph cluster")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'osdmap').get(
            'osdmap').get(
            'num_osds') == num_osds_after_patch,
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="OSDs number doesn't match expected number after {} sec. "
                    "Current osd number is {}. But shoud be {}".format(
            ceph_health_timeout,
            yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
                'osdmap').get(
                'osdmap').get(
                'num_osds'), num_osds_after_patch))

    # Wait for all osds are UP in cluster
    LOG.info("Waiting for all OSDs are UP in ceph cluster")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'osdmap').get(
            'osdmap').get(
            'num_up_osds') == num_osds_after_patch,
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Some osds are not UP after {} sec. "
                    "Current osds UP {}. But shoud be {}".format(
            ceph_health_timeout,
            yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
                'osdmap').get(
                'osdmap').get(
                'num_up_osds'), num_osds_after_patch))

    # Wait for HEALTH_OK after osds were added
    LOG.info("Final check ceph cluster status")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get(
            'status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))
    LOG.info("Ceph cluster is healthy. Congratulations!")

    osd_tree_data = yaml.safe_load(ceph_tools_pod.exec(cmd_osd_tree))
    osds = [node for node in osd_tree_data.get('nodes') if
            node.get('type') == 'osd']
    osds_dict = {}
    for osd in osds:
        osds_dict[osd.get('name')] = {'status': osd.get('status'),
                                      'reweight': osd.get('reweight')}
    LOG.info("Ceph osds info:{}{}".format(
        "\n", yaml.dump(osds_dict)))
