import pytest
import yaml
import base64

from si_tests import settings
from si_tests import logger
from si_tests.utils import waiters, templates

LOG = logger.logger

cluster_name = settings.TARGET_CLUSTER
namespace_name = settings.TARGET_NAMESPACE
region = settings.TARGET_REGION
deploy_image_url = settings.IRONIC_DEPLOY_IMAGE_URL
deploy_image_name = settings.TEMPEST_CUSTOM_IMAGE
bm_flavor = settings.TEMPEST_CUSTOM_FLAVOR


@pytest.mark.usefixtures('log_method_time')
def test_add_ironic_node(kaas_manager, show_step):
    """Add bm compute node.

    Scenario:
        1. Gather nodes information from yaml file
        2. Get openstack-client pod
        3. Get images UUIDs
        4. Delete nodes before creating
        5. Get nodes from child data and start inspecting
        6. Wait for all nodes stay available
    """
    LOG.info(f"Namespace: {namespace_name}; Cluster: {cluster_name}")
    ns = kaas_manager.get_namespace(namespace_name)
    cluster = ns.get_cluster(cluster_name)

    show_step(1)
    child_data = yaml.safe_load(templates.render_template(
        settings.BM_CHILD_SETTINGS_YAML_PATH))[0]
    bm_hosts_data = child_data['ironic_nodes']
    show_step(2)
    client_pod = cluster.k8sclient.pods.list(
        namespace="openstack",
        name_prefix='keystone-client')
    assert len(client_pod) > 0, ("No pods found with prefix "
                                 "keystone-client in namespace "
                                 "openstack")
    client_pod = client_pod[0]

    LOG.info("WA FOR SEGMENTATION PRODX-14122(PRODX-12461)."
             " SHOULD BE FIXED IN 21.3")
    client_pod.exec(['/bin/sh', '-c',
                     'openstack network set --provider-segment '
                     '470 ironic-provisioning'])

    show_step(3)
    LOG.info("Get images UUIDs")
    iku_cmd = ['/bin/sh', '-c', 'openstack image show -f '
               'value -c id ironic-agent.kernel']
    image_kernel_uuid = client_pod.exec(iku_cmd).strip()
    iru_cmd = ['/bin/sh', '-c', 'openstack image show -f '
               'value -c id ironic-agent.initramfs']
    image_ramdisk_uuid = client_pod.exec(iru_cmd).strip()

    show_step(4)
    resource_class_normalized = ''
    ironic_nodes = []

    # Force delete node before creating
    delete_nodes = [n['name'] for n in bm_hosts_data
                    if 'ironic' in n.get('si_roles', [])]
    for _node in delete_nodes:
        client_pod.exec(['/bin/sh', '-c',
                         'openstack baremetal node maintenance set'
                         f" {_node} &&"
                         f'openstack baremetal node delete {_node}'])
    cmd_ironic_nodes_list = ['/bin/sh', '-c',
                             'openstack baremetal node list -f yaml '
                             '--field uuid provision_state']

    def check_deleted(count=None):
        nodes_count = client_pod.exec(cmd_ironic_nodes_list)
        bm_list = yaml.safe_load(nodes_count)
        bm_stat = len(bm_list)
        LOG.info(f"Count of existing BM: {bm_stat}")
        if bm_stat == count and len(bm_list) == count:
            return True
        else:
            return False
    waiters.wait(
        lambda: check_deleted(count=0),
        timeout=900, interval=30,
        timeout_msg="Some bm nodes is not deleted")

    show_step(5)
    for _node in bm_hosts_data:
        if 'ironic' in _node.get('si_roles', []):
            ironic_nodes.append(_node)
            ironic_bm_node_name = _node.get('name')
            ipmi_ip = _node['ipmi']['ipmi_ip']
            ipmi_user = base64.b64decode(
                _node['ipmi']['username']).decode("utf-8")
            ipmi_password = base64.b64decode(
                _node['ipmi']['password']).decode("utf-8")
            pxe_mac = _node['networks'][0]['mac']
            node_mem_mb = _node['ironic_specific']['mem']
            node_cpus = _node['ironic_specific']['cpus']
            node_disk_gb = _node['ironic_specific']['disk']
            physnet = _node['ironic_specific']['physnet']
            multitenantcy = _node.get('ironic_specific')\
                .get('mutitenantcy').get('enabled')
            mt_params = ''
            mt_params_port = ''
            if multitenantcy:
                LOG.info(f"Multitenantcy enabled for <{ironic_bm_node_name}>")
                switch_name = _node['ironic_specific']['mutitenantcy'][
                    'switch_name']
                switch_port = _node['ironic_specific']['mutitenantcy'][
                    'switch_port']
                mt_params = ' --network-interface neutron'
                mt_params_port = ' --local-link-connection'\
                                 f' switch_info={switch_name}'\
                                 ' --local-link-connection '\
                                 'switch_id="FE:FE:FE:FE:FE:FE"'\
                                 ' --local-link-connection '\
                                 f'port_id={switch_port}'
            resource_class = ('ram' + str(node_mem_mb) + '_cpus'
                              + str(node_cpus) + '_disk'
                              + str(node_disk_gb)).upper()
            resource_class_normalized = resource_class.upper()
            cmd_add_bm = ['/bin/sh', '-c',
                          'openstack baremetal node create'
                          ' --driver ipmi'
                          ' --property cpu_arch=x86_64'
                          + mt_params +
                          f' --property memory_mb={node_mem_mb}'
                          f' --property cpus={node_cpus}'
                          f' --property local_gb={node_disk_gb}'
                          f' --driver-info deploy_kernel={image_kernel_uuid}'
                          f' --driver-info deploy_ramdisk={image_ramdisk_uuid}'
                          f' --driver-info ipmi_address={ipmi_ip}'
                          f' --driver-info ipmi_username={ipmi_user}'
                          f' --driver-info ipmi_password={ipmi_password}'
                          f' --resource-class {resource_class}'
                          f' --name {ironic_bm_node_name} -f yaml']
            LOG.info(f"Create baremetal node <{ironic_bm_node_name}>. "
                     f"Command:\n {cmd_add_bm}\n")
            exec_result = client_pod.exec(cmd_add_bm)
            LOG.debug(exec_result)
            LOG.info("Get created UUID")
            node_uuid = yaml.safe_load(exec_result).get('uuid')

            cmd_add_port = ['/bin/sh', '-c',
                            f'openstack baremetal port create'
                            f' {pxe_mac}'
                            f' --node {node_uuid}'
                            f' --physical-network {physnet}'
                            + mt_params_port]
            LOG.info(f"Create baremetal port for <{ironic_bm_node_name}>")
            client_pod.exec(cmd_add_port)
            _node.update({'uuid': node_uuid})
            cmd_provide = ['/bin/sh', '-c',
                           f'openstack baremetal node manage'
                           f' {node_uuid} && '
                           f'openstack baremetal node provide {node_uuid}']
            client_pod.exec(cmd_provide)
            LOG.info("Provide baremetal node")

    assert len(ironic_nodes) >= 1, "BMH with si_roles 'ironic' not found" \
                                   " in child data file"

    show_step(6)
    cmd_ironic_nodes_list = ['/bin/sh', '-c',
                             'openstack baremetal node list -f yaml '
                             '--field uuid provision_state']
    exec_node_count = client_pod.exec(cmd_ironic_nodes_list)
    LOG.info(f"Get count of nodes. Command result: {exec_node_count}")
    ironic_nodes_count = len(yaml.safe_load(exec_node_count))

    def check(count=None):
        nodes_count = client_pod.exec(cmd_ironic_nodes_list)
        bm_list = yaml.safe_load(nodes_count)
        bm_stat = sum(h.get('Provisioning State') == 'available'
                      for h in bm_list)
        LOG.info(f"Count ready BM: {bm_stat} "
                 f"expected {ironic_nodes_count}")
        if bm_stat == count and len(bm_list) == count:
            return True
        else:
            return False

    waiters.wait(
        lambda: check(count=ironic_nodes_count),
        timeout=900, interval=30,
        timeout_msg="Some bm nodes is not provided by ironic")

    LOG.info("Download and create image")
    client_pod.exec(['/bin/sh', '-c',
                     f'wget -O /tmp/i {deploy_image_url} &&'
                     f'openstack image create {deploy_image_name} --public '
                     '--disk-format qcow2 --container-format bare '
                     '--file /tmp/i'], stderr=True)

    LOG.info(f"Create baremetal flavor <{bm_flavor}>")

    client_pod.exec(['/bin/sh', '-c',
                     f'openstack flavor create {bm_flavor} --property '
                     'resources:VCPU=0 --property resources:MEMORY_MB=0 '
                     '--property resources:DISK_GB=0 --property '
                     'resources:CUSTOM_BAREMETAL_'
                     f'{resource_class_normalized}=1 --disk 1'], stderr=True)
