import time
import yaml
import exec_helpers

from kubernetes import client as kube_client
from si_tests import settings
from si_tests.utils import waiters, utils
from si_tests import logger
from si_tests.clients.k8s import K8sCluster

LOG = logger.logger


def test_remove_ceph_osd_node():
    """This test removes osd node from already created ceph cluster.

       1. Get information about ceph cluster using ceph tools pod.
       2. Define node for downscaling according to SCALE_NODE_NAME_PATTERN
       3. Get osds belonging to that node
       4. Making osds out from cluster
       5. Delete appropriate deployments
       6. Purging osds
       7. Remove node data from rook-ceph and cephclusters CRDs
       8. Remove node from ceph cluster
       9. Restart osd prepare jobs
       10. Clean node from ceph dependencies (To get node be ready for other,
           not related to ceph actions)

       Parameters required for test execution:
         - KUBECONFIG
         - SCALE_NODE_NAME_PATTERN
         - ROOK_CEPH_NS
         - CEPH_LCM_MIRANTIS_NS
         - MOSK_PRIVATE_KEY_PATH
    """

    kubeconfig = settings.KUBECONFIG_PATH
    scale_node_name_pattern = settings.SCALE_NODE_NAME_PATTERN
    rook_ns = settings.ROOK_CEPH_NS
    ceph_lcm_mirantis_ns = settings.CEPH_LCM_MIRANTIS_NS
    private_key = settings.MOSK_PRIVATE_KEY_PATH
    assert private_key, "Private key path is not defined"
    ceph_health_timeout = 2400
    # Init k8s client
    client = K8sCluster(kubeconfig=kubeconfig)
    # https://github.com/kubernetes-client/python/issues/234
    delete_options = kube_client.V1DeleteOptions(
        api_version='v1',
        kind="DeleteOptions",
        propagation_policy="Background")

    # Getting pods for checking Ceph status
    ceph_tools_pod = client.pods.list(namespace=rook_ns,
                                      name_prefix='rook-ceph-tools')
    assert len(ceph_tools_pod) > 0, ("No pods found with prefix "
                                     "rook-ceph-tools in namespace {}".format(
                                         rook_ns))

    ceph_tools_pod = ceph_tools_pod[0]
    cmd_status = ['/bin/sh', '-c', 'ceph -s -f json']
    ceph_devices_cmd = ['/bin/sh', '-c', 'ceph device ls -f json']

    # Determine node for downscale
    scale_node = [node for node in
                  client.nodes.list_all()
                  if scale_node_name_pattern in node.name]
    assert len(scale_node) > 0, ("No nodes with name pattern {} "
                                 "found for scaling.".format(
                                     scale_node_name_pattern))
    scale_node = scale_node[0]
    scale_node_data = scale_node.read().to_dict()
    scale_node_ip = [ip.get(
        'address') for ip in scale_node_data.get('status').get(
        'addresses') if 'InternalIP' in ip.get('type')]
    assert len(scale_node_ip) > 0, ("No IPs found for type InternalIP "
                                    "in node addresses. Current IP "
                                    "information: {}".format(
                                        scale_node_data.get(
                                            'status').get('addresses', [])))
    scale_node_ip = scale_node_ip[0]
    scale_node_name = scale_node.name
    LOG.info("Next node will be removed from ceph cluster: {}".format(
        scale_node_name))

    all_devices = yaml.safe_load(ceph_tools_pod.exec(ceph_devices_cmd))
    scale_node_devices = [device.get('location')[0].get('dev') for device
                          in all_devices if scale_node_name in device.get(
                              'location')[0].get('host')]
    node_osds = [device.get('daemons', []) for device in all_devices if
                 scale_node_name in device.get('location')[0].get('host')]
    osds_list = [osd for osds in node_osds for osd in osds]
    LOG.info("Check if HEALTH_OK before patching CRD")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))

    # Get Active+Clean pages to compare after purging
    pages = yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get('pgmap')
    active_clean_pages = [pg.get('count') for pg in pages.get(
        'pgs_by_state') if 'active+clean' in pg.get('state_name')][0]

    for osd in osds_list:
        ceph_osd_out_cmd = ['/bin/sh', '-c', 'ceph osd out {}'.format(osd)]
        ceph_tools_pod.exec(ceph_osd_out_cmd)
        LOG.info("Marked {} out".format(osd))

    LOG.info('Check for outing progress')
    progress = yaml.safe_load(ceph_tools_pod.exec(
        cmd_status)).get('progress_events', {})
    if progress:
        LOG.info(progress)
        LOG.info("Waiting for progress finished.")
        waiters.wait(
            lambda: not yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
                'progress_events'), timeout=ceph_health_timeout, interval=30,
            timeout_msg="Failed to finish outing osds in {} sec. "
                        "Current status: {}".format(ceph_health_timeout,
                                                    yaml.safe_load(
                                                        ceph_tools_pod.exec(
                                                            cmd_status)).get(
                                                        'progress_events',
                                                        {})))

    LOG.info("Check osds info after osds out")
    stdout = ceph_tools_pod.exec(['/bin/sh', '-c', 'ceph osd tree'])
    LOG.info("\n {}".format(stdout))

    LOG.info("Delete appropriate deployments")
    # Delete deployment according to osd ID
    for osd in osds_list:
        osd_id_number = osd.split('.')[1]
        deployment = client.deployments.get(
            name="rook-ceph-osd-{}".format(osd_id_number), namespace=rook_ns)
        LOG.info("Deleting deployment rook-ceph-osd-{}".format(osd_id_number))
        deployment.delete()

    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))

    LOG.info("Check osds info after deployments are deleted")
    stdout = ceph_tools_pod.exec(['/bin/sh', '-c', 'ceph osd tree'])
    LOG.info("\n {}".format(stdout))

    for osd in osds_list:
        purge_cmd = ['/bin/sh', '-c', 'ceph osd purge {}'.format(osd)]
        LOG.info("Purging {}".format(osd))
        stdout = ceph_tools_pod.exec(purge_cmd).strip()
        LOG.info(stdout)
        # Need to wait few sec for data refreshed
        time.sleep(5)
        LOG.info("Waiting for rebalance after {} is purged".format(osd))

        # Due to after purging osds health is OK, but pages are being remapped,
        # we need to compare active+clean pages count, not health status.
        waiters.wait(
            lambda: [pg.get('count') for pg in yaml.safe_load(
                ceph_tools_pod.exec(cmd_status)).get('pgmap').get(
                    'pgs_by_state') if 'active+clean' in pg.get(
                        'state_name')][0] == active_clean_pages,
            timeout=ceph_health_timeout, interval=10,
            timeout_msg="Some pages are not active after {} sec".format(
                ceph_health_timeout))
    LOG.info("Check osds info after osds are purged")
    stdout = ceph_tools_pod.exec(['/bin/sh', '-c', 'ceph osd tree'])
    LOG.info("\n {}".format(stdout))
    miraceph_crd = [crd for crd in client.miracephs.list(
        namespace=ceph_lcm_mirantis_ns) if crd.name == 'cephcluster']
    assert miraceph_crd, ("CRD with name cephcluster is not found")
    miraceph_crd = miraceph_crd[0]
    new_crd = miraceph_crd.read()
    new_crd.spec.get('nodes').pop(scale_node_name)
    miraceph_crd.replace(body=new_crd)
    rook_crd = [crd for crd in client.rookcephclusters.list_all()
                if crd.name == 'rook-ceph']
    assert rook_crd, ("CRD with name rook-ceph is not found")
    rook_crd = rook_crd[0]
    rook_crd_for_patch = rook_crd.read()
    nodes = rook_crd.data.get('spec').get('storage').get('nodes')
    new_nodes = [node for node in nodes if
                 scale_node_name not in node.get('name')]
    rook_crd_for_patch.spec['storage']['nodes'] = new_nodes
    # Check if HEALTH_OK before patching
    LOG.info("Check if HEALTH_OK before patching CRD")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))
    rook_crd.replace(body=rook_crd_for_patch)

    # Remove node from ceph custer
    remove_node_cmd = ['/bin/sh', '-c', 'ceph osd crush rm {}'.format(
        scale_node_name)]
    ceph_tools_pod.exec(remove_node_cmd)
    # Restart osd prepare jobs
    osd_prepare_jobs = client.jobs.list(
        namespace=rook_ns, name_prefix='rook-ceph-osd-prepare')
    for job in osd_prepare_jobs:
        client.api_batch.delete_namespaced_job(
            namespace=rook_ns, name=job.name, body=delete_options)

    LOG.info("Final check osds tree and ceph status")
    ceph_stdout = ceph_tools_pod.exec(['/bin/sh', '-c', 'ceph -s'])
    osd_tree_stdout = ceph_tools_pod.exec(['/bin/sh', '-c', 'ceph osd tree'])
    LOG.info("\n {}".format(ceph_stdout))
    LOG.info("\n {}".format(osd_tree_stdout))

    # Clean node (Additional step, to get node be ready for other actions)

    LOG.info("Cleaning node {} and reboot".format(scale_node_name))
    keys = utils.load_keyfile(private_key)
    pkey = utils.get_rsa_key(keys['private'])
    auth = exec_helpers.SSHAuth(username='ubuntu', password='', key=pkey)
    ssh = exec_helpers.SSHClient(
        host=scale_node_ip, port=22, auth=auth)
    ssh.logger.addHandler(logger.console)
    ssh.sudo_mode = True
    for device in scale_node_devices:
        ssh.check_call(
            "dd if=/dev/zero of=/dev/{} bs=100M count=10".format(device))

    # Remove Ceph directory
    ssh.check_call("rm -rf /var/lib/rook/")

    # Reboot node. No need to wait until UP
    ssh.check_call("/sbin/shutdown -r -f now &>/dev/null & exit")
    LOG.info("Done")
