import yaml
import exec_helpers

from kubernetes import client as kube_client
from si_tests import settings
from si_tests.utils import waiters, utils
from si_tests import logger
from si_tests.clients.k8s import K8sCluster

LOG = logger.logger


def test_replace_failed_osd_disk():
    """This test emulates disk crashing and replacing

       1. Get information about ceph cluster using ceph tools pod.
       2. Define node and disk for tests
       3. Emulate crash
       4. Check required signals (WARNING,RECENT_CRASH,OSD_DOWN)
       4. Make all required steps according do docs
       4. Verify that ceph cluster is healthy after all operations

       Parameters required for test execution:
         - KUBECONFIG
         - FAILED_DISK_NAME
         - FAILED_DISK_NODE_NAME_PATTERN
         - ROOK_CEPH_NS
         - MOSK_PRIVATE_KEY_PATH
    """

    failed_disk_name = settings.FAILED_DISK_NAME
    failed_disk_node_name_pattern = settings.FAILED_DISK_NODE_NAME_PATTERN
    rook_ns = settings.ROOK_CEPH_NS
    private_key = settings.MOSK_PRIVATE_KEY_PATH
    assert private_key, "Private key path is not defined"

    # Init k8s client
    client = K8sCluster(kubeconfig=settings.KUBECONFIG_PATH)
    # https://github.com/kubernetes-client/python/issues/234
    delete_options = kube_client.V1DeleteOptions(
        api_version='v1',
        kind="DeleteOptions",
        propagation_policy="Background")

    # Getting pods for checking Ceph status
    ceph_tools_pod = client.pods.list(namespace=rook_ns,
                                      name_prefix='rook-ceph-tools')
    assert len(ceph_tools_pod) > 0, ("No pods found with prefix "
                                     "rook-ceph-tools in namespace {}".format(
                                        rook_ns))

    ceph_operator_pod = client.pods.list(namespace=rook_ns,
                                         name_prefix='rook-ceph-operator')
    assert len(ceph_operator_pod) > 0, ("No pods found with prefix "
                                        "rook-ceph-operator in "
                                        "namespace {}".format(rook_ns))
    ceph_tools_pod = ceph_tools_pod[0]
    ceph_operator_pod = ceph_operator_pod[0]

    cmd_status = ['/bin/sh', '-c', 'ceph -s -f json']
    cmd_osd_tree = ['/bin/sh', '-c', 'ceph osd tree -f json']
    archive_cmd = ['/bin/sh', '-c', 'ceph crash archive-all']

    # Determine node for working with disk
    failed_disk_node = [node for node in client.nodes.list_all()
                        if failed_disk_node_name_pattern in node.name]
    assert len(failed_disk_node) > 0, ("No nodes with name pattern {} "
                                       "found.".format(failed_disk_node))

    failed_disk_node = failed_disk_node[0]
    failed_disk_node_name = failed_disk_node.name
    failed_disk_node_data = failed_disk_node.read().to_dict()
    LOG.info("Working with disk {disk} on node {node}.".format(
        disk=failed_disk_name, node=failed_disk_node_name))
    failed_disk_node_ip = [ip.get(
        'address') for ip in failed_disk_node_data.get('status').get(
        'addresses') if 'InternalIP' in ip.get('type')]
    assert len(failed_disk_node_ip) > 0, ("No IPs found for type InternalIP "
                                          "in node addresses. Current IP "
                                          "information: {}".format(
                                              failed_disk_node_data.get(
                                                  'status').get(
                                                      'addresses', [])))
    failed_disk_node_ip = failed_disk_node_ip[0]
    ceph_health_timeout = 2400
    keys = utils.load_keyfile(private_key)
    pkey = utils.get_rsa_key(keys['private'])
    auth = exec_helpers.SSHAuth(username='ubuntu',
                                password='', key=pkey)
    ssh = exec_helpers.SSHClient(
            host=failed_disk_node_ip, port=22, auth=auth)
    ssh.logger.addHandler(logger.console)

    LOG.info("Verify that all osds are UP")
    osds_info = yaml.safe_load(ceph_tools_pod.exec(cmd_osd_tree))
    osds = [node for node in osds_info.get('nodes') if
            node.get('type') == 'osd']
    nodes = [node for node in osds_info.get('nodes') if
             node.get('type') == 'host']
    osds_dict = {}
    nodes_dict = {}
    for osd in osds:
        osds_dict[osd.get('name')] = {'status': osd.get('status'),
                                      'reweight': osd.get('reweight')}
    for node in nodes:
        nodes_dict[node.get('name')] = {'osds': [
            'osd.' + str(child) for child in node.get('children')]}
    LOG.info("Ceph nodes info:{}{}".format("\n", yaml.dump(nodes_dict)))
    LOG.info("Ceph osds info:{}{}".format("\n", yaml.dump(osds_dict)))

    # Here we will simulate disk crash
    LOG.info("Cleaning {disk} disk on node {node} "
             "and remove it from system".format(
                 disk=failed_disk_name, node=failed_disk_node_name))
    ssh.sudo_mode = True
    ssh.check_call(
        "dd if=/dev/zero of=/dev/{} bs=100M count=10".format(
            failed_disk_name))
    ssh.check_call(
        "echo '1' > /sys/block/{}/device/delete".format(
            failed_disk_name))

    # Wait for HEALTH_WARN
    LOG.info("Waiting for WARNING signal after disk was crashed.")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_WARN",
        timeout=ceph_health_timeout, interval=30)

    # Check that only 1 osd is down and belongs to choosen node
    osds_info = yaml.safe_load(ceph_tools_pod.exec(cmd_osd_tree))
    osds_down = [node.get('name') for node in osds_info.get(
        'nodes') if node.get('type') == 'osd' and node.get(
        'status') == 'down']
    assert len(osds_down) == 1, ("More that 1 osds are down: {}".format(
        osds_down))
    osds = [node for node in osds_info.get('nodes') if
            node.get('type') == 'osd']
    for osd in osds:
        osds_dict[osd.get('name')] = {'status': osd.get('status'),
                                      'reweight': osd.get('reweight')}
    LOG.info("Current osds stats: {}{}".format("\n", yaml.dump(osds_dict)))

    # Now we need to wait until information about crash appears in ceph status
    LOG.info("Waiting for information about "
             "recent crash appears in ceph status")
    waiters.wait(
        lambda: 'RECENT_CRASH' in yaml.safe_load(
            ceph_tools_pod.exec(cmd_status)).get('health').get(
            'checks'), timeout=ceph_health_timeout, interval=30)

    LOG.info("Crash INFO: {}".format(yaml.safe_load(ceph_tools_pod.exec(
        cmd_status)).get('health').get('checks')))

    LOG.info("Archiving information about crash")
    ceph_tools_pod.exec(archive_cmd)

    # Wait for rebalancing and HEALTH_OK.
    LOG.info("Waiting for data rebalanced and HEALTH is OK")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))

    osd_tree_data = yaml.safe_load(ceph_tools_pod.exec(cmd_osd_tree))

    # Define osd num which is down after disk crash
    LOG.info("Define osd ID for further work")
    osd_num = osds_down[0]
    LOG.info("OSD ID is {}".format(osd_num))
    osd_id_number = osd_num.split('.')[1]
    purge_cmd = ['/bin/sh', '-c', 'ceph osd purge {}'.format(
        osd_num)]
    LOG.info("Purging {}".format(osd_num))
    ceph_tools_pod.exec(purge_cmd)

    # Delete deployment according to osd ID
    deployment = client.deployments.get(
        name="rook-ceph-osd-{}".format(osd_id_number), namespace=rook_ns)
    LOG.info("Deleting deployment rook-ceph-osd-{}".format(osd_id_number))
    deployment.delete()
    LOG.info("Done")

    # Wait for rebalancing
    LOG.info("Waiting for data rebalanced and HEALTH is OK")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))

    LOG.info("Rebooting node {}".format(failed_disk_node_name))
    ssh.check_call("rm -rf /var/lib/rook/{}".format(osd_num))
    ssh.check_call("iptables -A INPUT -p tcp --dport 22 -j REJECT; "
                   "/sbin/shutdown -r -f now &>/dev/null & exit")
    LOG.info("Node is being rebooted.")
    waiters.wait_tcp(failed_disk_node_ip, port=22, timeout=600)
    LOG.info("Node is rebooted succesfully. Establishing new connection...")
    ssh = exec_helpers.SSHClient(host=failed_disk_node_ip, port=22, auth=auth)
    ssh.logger.addHandler(logger.console)

    LOG.info("Restarting ceph-operator pod and osd-prepare jobs")
    ceph_operator_pod.delete()
    osd_prepare_jobs = client.jobs.list(
        namespace=rook_ns, name_prefix='rook-ceph-osd-prepare')
    for job in osd_prepare_jobs:
        client.api_batch.delete_namespaced_job(
            namespace=rook_ns, name=job.name, body=delete_options)

    # Wait for new osd appears in cluster
    LOG.info("Waiting for {} appears in ceph cluster".format(
        osd_num))
    waiters.wait(
        lambda: osd_num in [osd.get('name') for osd in yaml.safe_load(
            ceph_tools_pod.exec(cmd_osd_tree)).get('nodes')],
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="{} is not existed in ceph cluster "
                    "after {} sec".format(osd_num, ceph_health_timeout))

    LOG.info("OSD {} existed in ceph cluster. Continue".format(osd_num))
    LOG.info("Waiting for {} become UP status".format(osd_num))
    waiters.wait(
        lambda: [node.get('status') for node in yaml.safe_load(
            ceph_tools_pod.exec(cmd_osd_tree)).get('nodes') if node.get(
                'name') == osd_num][0] == "up",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="{} is failed to bacome UP status "
                    "after {} sec".format(osd_num, ceph_health_timeout))

    LOG.info("Waiting for data rebalanced and HEALTH is OK")
    waiters.wait(
        lambda: yaml.safe_load(ceph_tools_pod.exec(cmd_status)).get(
            'health').get('status') == "HEALTH_OK",
        timeout=ceph_health_timeout, interval=30,
        timeout_msg="Ceph status failed to became "
                    "HEALTH_OK in {} sec".format(ceph_health_timeout))

    LOG.info("Everything seems to be fine. Congratulations!")
    osd_tree_data = yaml.safe_load(ceph_tools_pod.exec(cmd_osd_tree))
    osds = [node for node in osd_tree_data.get('nodes') if
            node.get('type') == 'osd']
    osds_dict = {}
    for osd in osds:
        osds_dict[osd.get('name')] = {'status': osd.get('status'),
                                      'reweight': osd.get('reweight')}
    LOG.info("Ceph osds info:{}{}".format("\n", yaml.dump(osds_dict)))
