import pytest
import os
import re
import yaml

import kubernetes

from si_tests.deployments import openstack_deploy
from si_tests.managers import clustercheck_manager, openstack_manager
from si_tests import settings
from si_tests import logger

LOG = logger.logger


@pytest.mark.parametrize("_", ["CLUSTER_NAME={0}"
                               .format(settings.TARGET_CLUSTER)])
@pytest.mark.usefixtures("store_cluster_description")
@pytest.mark.usefixtures("introspect_PRODX_9696")
def test_update_child_clusterrelease(kaas_manager, _):
    """Update child cluster release"""

    cluster_name = settings.TARGET_CLUSTER
    namespace_name = settings.TARGET_NAMESPACE
    update_release_name = settings.KAAS_CHILD_CLUSTER_UPDATE_RELEASE_NAME

    ns = kaas_manager.get_namespace(namespace_name)
    child_cluster = ns.get_cluster(cluster_name)
    if not update_release_name:
        LOG.info("KAAS_CHILD_CLUSTER_UPDATE_RELEASE_NAME wasn't set, trying "
                 "to acquire available clusterrelease version for upgrade")
        cr_available = child_cluster.available_clusterrelease_version
        assert cr_available is not None, "Version for upgrade wasn't found"
        LOG.info("Cluster release version {0} will be used for upgrade"
                 .format(cr_available))
        update_release_name = cr_available
    cr_before = child_cluster.clusterrelease_version
    if update_release_name == cr_before:
        pytest.skip(f"Requested {update_release_name} is the same as current "
                    f"clusterrelease version {cr_before}")
    LOG.info("Collect child cluster fixed resources before update")
    child_client = ns.get_cluster(settings.TARGET_CLUSTER)
    lcm_type = child_cluster.data['status'].get(
        'providerStatus', {}).get('releaseRefs', {}).get(
        'current', {}).get('lcmType', '')
    LOG.info("LCM Type is {0}".format(lcm_type))
    cluster_resources_before = child_client.describe_fixed_resources()
    LOG.debug("Child cluster resources before update:\n{0}"
              .format(yaml.dump(cluster_resources_before)))
    ucp_tag_in_cr_before = set(
        [x['params']['ucp_tag'] for x in kaas_manager.get_clusterrelease(
            cr_before).data['spec']['machineTypes']['control']
         if 'ucp_tag' in x['params'].keys()])

    LOG.info("Cluster release before update {0}".format(cr_before))
    LOG.info('ucp_tag before update {0}'.format(ucp_tag_in_cr_before))
    LOG.info(("=" * 80) + "\n\nUpdating child cluster {0} "
             "to the clusterrelease {1}\n\n"
             .format(cluster_name, update_release_name))

    if re.search(r'mos-6-1[234]-.?', child_cluster.clusterrelease_version):
        LOG.info("Apply WA PRODX-12511: Set pdb 1 for horzion")
        # TODO (tleontovich) Delete WR after fix PRODX-12511
        try:
            horizon_pdb = child_cluster.k8sclient.pdb.get(
                'horizon', 'openstack')
            horizon_pdb.patch({"spec": {"minAvailable": 1}})
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                LOG.info("There is no MOS apps deployed.")

    child_cluster.update_cluster(update_release_name)
    child_cluster.check.check_cluster_release(update_release_name)
    child_cluster.check.check_update_finished()
    # Nodes will not have updating status if there is no changes
    # in ucp_tag or CR versions
    ucp_tag_after = set(
        [x['params']['ucp_tag']
         for x in kaas_manager.get_clusterrelease(
            update_release_name).data['spec']['machineTypes']['control']
         if 'ucp_tag' in x['params'].keys()]
    )
    child_cluster.check.check_cluster_nodes()
    child_cluster.check.check_helmbundles()
    child_cluster.check.check_k8s_nodes()
    # Check/wait for correct docker service replicas in cluster
    # We have renamed agent if ucp changes version
    if update_release_name != cr_before and \
            ucp_tag_in_cr_before != ucp_tag_after:
        child_cluster.check.check_actual_expected_docker_services(
            changed_after_upd={'ucp-worker-agent-x': 'ucp-worker-agent-y'})
    else:
        child_cluster.check.check_actual_expected_docker_services()
        assert ucp_tag_in_cr_before == ucp_tag_after,\
            "UCP tag was updated, when shouldn't"

    child_cluster.check.check_k8s_pods()
    child_cluster.check.check_actual_expected_pods(timeout=3200)
    if 'mos' in settings.KAAS_CHILD_CLUSTER_RELEASE_NAME:
        expected_fails = clustercheck_manager.EXCLUDED_SL_COMPONENTS
        child_cluster.check.\
            check_cluster_readiness(exp_provider_status=False,
                                    expected_fails=expected_fails)
    elif 'mke' in settings.KAAS_CHILD_CLUSTER_RELEASE_NAME:
        child_cluster.check.check_cluster_readiness()

    # check k8s version
    failed = {}
    for node in child_cluster.k8sclient.nodes.list_all():
        node_info = node.read().status.node_info
        if node_info.kubelet_version != node_info.kube_proxy_version:
            failed[node.name] = "kubelet_version {0} doesn't match " \
                                "kube_proxy_version {1} version".format(
                node_info.kubelet_version,
                node_info.kube_proxy_version)

    assert failed == {}, "k8s versions mismatch " \
                         "Details: {}".format(failed)

    child_cluster.store_k8s_artifacts()
    child_cluster.provider_resources.save_artifact()

    # Check for changed/missing resources
    child_cluster.check.check_k8s_resources_after_update(
        child_client, cluster_resources_before)
    LOG.info("\n*** Child cluster <{0}> "
             "have been upgraded to the clusterrelease {1}"
             .format(cluster_name, update_release_name))

    if re.search(r"mos.?-.*", child_cluster.clusterrelease_version):
        LOG.info("Child cluster name contains MOS pattern. "
                 "coredns ConfigMap is going to be adjusted.")
        child_kubeconfig_name, child_kubeconfig = \
            child_cluster.get_kubeconfig_from_secret()
        child_kubeconfig_path = os.path.join(
            settings.ARTIFACTS_DIR, child_kubeconfig_name)
        os_manager = openstack_manager.OpenStackManager(
            kubeconfig=child_kubeconfig_path)
        timeouts = settings.OPENSTACK_DEPLOY_TIMEOUT
        try:
            os_manager.get_openstackdeployment(
                name=settings.OSH_DEPLOYMENT_NAME, read=True)
        except kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                LOG.info("There is no MOS apps deployed.")
        else:
            LOG.info("Configuring coredns in kube-system to work "
                     "with it.just.works domain")
            openstack_deploy.configure_dns_kubesys(
                os_manager,
                timeouts.get("3", 180)
            )
            os_manager.wait_osdpl_status(timeout=3600)
            LOG.info("Wait osdpl health status=Ready")
            os_manager.wait_openstackdeployment_health_status(
                timeout=1800)
            LOG.info("Wait os jobs to success and pods to become Ready")
            os_manager.wait_os_resources(timeout=1800)
